#!/usr/bin/env python

import os
import sys
import numpy as np

try:
    import matplotlib
    matplotlib.use("Agg")
    import matplotlib.pyplot as plt
except BaseException:
    pass

EPSILON = 1e-12


def scale_var(var, min_range, max_range):
    return (var - min_range) / max((max_range - min_range), EPSILON)


def unscale_var(var, min_range, max_range):
    return (var * (max_range - min_range)) + min_range


def visualize_fitness(fitness_file=None, best_fitnesses=None,
                      mean_fitnesses=None, outfile="fitness.png"):
    if best_fitnesses is None or mean_fitnesses is None:
        assert os.path.exists(fitness_file)
        best_fitnesses = []
        mean_fitnesses = []
        with open(fitness_file) as f:
            for line in f.readlines():
                best_fit, mean_fit = [float(x) for x in line.rstrip().split()]
                best_fitnesses.append(best_fit)
                mean_fitnesses.append(mean_fit)

        plt.figure(figsize=(16, 12))
        plt.plot(np.arange(len(best_fitnesses)), best_fitnesses,
                 label="Best Fitness")
        plt.plot(np.arange(len(mean_fitnesses)), mean_fitnesses,
                 label="Mean Fitness")
        plt.xlabel("Generation")
        plt.ylabel("Fitness")
        plt.grid()
        plt.legend(loc='lower right')

        outfile = os.path.abspath(os.path.expanduser(outfile))
        plt.savefig(outfile, bbox_inches='tight')


if __name__ == "__main__":
    func = eval(sys.argv[1])
    func(*sys.argv[2:])



import itertools
import math

import numpy as np


class Hyperparameter(object):
    def __init__(
            self,
            name,
            index=None,
            values=None,
            min_range=None,
            max_range=None,
            array_length=None,
            unbounded=None,
            is_log=None,
            rand_init=None):

        self.name = name
        self.index = index
        self.values = values
        self.min_range = min_range
        self.max_range = max_range
        self.unbounded = unbounded
        self.is_log = is_log
        self.rand_init = rand_init
        self.array_length = array_length

    def is_compatible(self, h):
        return self.index == h.index and \
            self.array_length == h.array_length

    def __str__(self):
        my_str = "Hyperparameter: %s " % self.name
        if self.index is not None:
            my_str += "Index: %s " % self.index
        if self.values is not None:
            my_str += "Value: %s " % self.values
        if self.min_range is not None:
            my_str += "Min range: %s " % self.min_range
        if self.max_range is not None:
            my_str += "Max range: %s " % self.max_range
        if self.array_length is not None:
            my_str += "Array length: %s " % self.array_length
        if self.unbounded is not None:
            my_str += "Unbounded: %s " % self.unbounded
        if self.is_log is not None:
            my_str += "Log scale: %s " % self.is_log
        if self.rand_init is not None:
            my_str += "Rand init: %s " % self.rand_init
        return my_str


class HyperparameterParser(object):
    '''Class for parsing hyperparameters'''

    def __init__(self, runner_args, logger):
        self.runner_args = runner_args
        self.logger = logger

    def convert_to_tuples(self, hyperparameters):
        if self.runner_args.optimizer == "grid":
            all_hyperparam_values = []
            for h in hyperparameters:
                assert h.values is not None
                hyperparam_values = [(h.name, value) for value in h.values]
                all_hyperparam_values.append(hyperparam_values)

            hyperparam_tuples = []
            for item in itertools.product(*all_hyperparam_values):
                hyperparam_tuple = {}
                for name, param in item:
                    hyperparam_tuple[name] = param
                hyperparam_tuples.append(hyperparam_tuple)
        else:
            hyperparam_tuples = []
            for hyperparam_list in hyperparameters:
                hyperparam_dict = {}
                for h in hyperparam_list:
                    hyperparam_dict[h.name] = h.values
                hyperparam_tuples.append(hyperparam_dict)

        # print hyperparam_tuples
        return hyperparam_tuples

    def parse(self):
        self.index = 0
        hyperparameters = []
        for hyperparam in self.runner_args.hyperparam:
            param_name = hyperparam.split('=')[0]
            param_values_str = hyperparam.split('=')[1]
            if self.runner_args.optimizer == "grid":
                hyperparameters.append(self._parse_grid(param_name,
                                                        param_values_str))
            else:
                hyperparameters.append(self._parse_opt(param_name,
                                                       param_values_str))
        if self.runner_args.verbose:
            self.logger.info("Parsed the following hyperparameters:")
            for h in hyperparameters:
                self.logger.info(str(h))
        return hyperparameters

    def _parse_opt(self, param_name, range_str):
        unbounded = is_log = rand_init = False
        min_range = max_range = array_length = None
        raw_fields = range_str.split(":")

        correct_format = True
        flags = ""
        if len(raw_fields) > 2:
            flags = raw_fields[-1]
            allowed_flags = 'ualr'
            for letter in flags:
                if letter not in allowed_flags or flags.count(letter) != 1:
                    correct_format = False
            proper_field_length = 4 if "a" in flags else 3
            if len(raw_fields) != proper_field_length:
                correct_format = False
        elif len(raw_fields) < 2:
            correct_format = False

        if not correct_format:
            raise ValueError("Hyperparameter flags (%s) are incorrect for %s" %
                             (range_str, self.runner_args.optimizer))

        try:
            min_range = float(raw_fields[0])
            max_range = float(raw_fields[1])
            array_length = int(raw_fields[2]) if "a" in flags else None
            if array_length is not None and array_length <= 0:
                raise ValueError
        except ValueError:
            raise ValueError(
                "Hyperparameter values (%s) are incorrect for %s" %
                (range_str, self.runner_args.optimizer))

        if min_range > max_range:
            raise ValueError("Min range (%s) is larger than max range (%s) " %
                             (min_range, max_range))

        unbounded = True if "u" in flags else False
        rand_init = True if "r" in flags else False
        if "l" in flags:
            is_log = True
            min_range = math.log(min_range)
            max_range = math.log(max_range)
        else:
            is_log = False

        if not hasattr(self, 'index'):
            self.index = 0
        h = Hyperparameter(
            param_name,
            index=self.index,
            min_range=min_range,
            max_range=max_range,
            array_length=array_length,
            unbounded=unbounded,
            is_log=is_log,
            rand_init=rand_init)
        self.index += array_length if array_length is not None else 1
        return h

    def _parse_grid(self, param_name, range_str):
        return_val = None
        if ',' in range_str:
            # return numpy array for consistency with other cases
            return_val = np.array([float(s) for s in range_str.split(',')])
        elif ':' in range_str:
            range_limits = range_str.split(':')
            assert len(range_limits) > 1
            if len(range_limits) == 2:
                try:
                    limit1 = float(range_limits[0])
                except ValueError:
                    limit1 = 0.0
                limit2 = float(range_limits[1])
                return_val = np.arange(limit1, limit2 + 1)
            else:
                try:
                    limit1 = float(range_limits[0])
                except ValueError:
                    limit1 = 0.0

                limit3 = float(range_limits[2])

                try:
                    limit2 = float(range_limits[1])
                    if int(limit2) == limit2 and limit2 > abs(limit3 - limit1):
                        return_val = np.linspace(limit1, limit3, int(limit2))
                    else:
                        return_val = np.arange(
                            limit1, limit3 + 0.5 * limit2, limit2)

                except ValueError:
                    if 'l' in range_limits[1]:
                        limit2 = int(range_limits[1].replace('l', ''))
                        return_val = np.exp(
                            np.linspace(
                                np.log(limit1),
                                np.log(limit3),
                                limit2))
                    else:
                        raise ValueError(
                            'unknown limit specification ' +
                            range_limits[1])

        else:
            return_val = [float(range_str)]
        if not isinstance(return_val, list):
            return_val = return_val.tolist()

        if not hasattr(self, 'index'):
            self.index = 0
        h = Hyperparameter(param_name, index=self.index, values=return_val)
        self.index += 1
        return h



import json
import time
import re
from .keyvalue_provider import KeyValueProvider
from .gcloud_artifact_store import GCloudArtifactStore
from .util import timeit


class GSProvider(KeyValueProvider):

    def __init__(self, config, blocking_auth=True, verbose=10, store=None):
        self.config = config
        self.bucket = config.get('bucket', 'studioml-meta')

        self.meta_store = GCloudArtifactStore(config, verbose)
        super(GSProvider, self).__init__(
            config,
            blocking_auth,
            verbose,
            store)

    @timeit
    def _get(self, key, shallow=False):
        bucket = self.meta_store._get_bucket_obj()
        retval = {}
        if shallow:
            blob_iterator = bucket.list_blobs(
                prefix=key, delimiter='/')
            bloblist = list(blob_iterator)
            blobnames = {b.name for b in bloblist}

            prefixes = blob_iterator.prefixes
            suffixes = [re.sub('^' + key, '', p) for p in prefixes | blobnames]

            retval = set({})
            for s in suffixes:
                if s.endswith('/'):
                    retval.add(s[:-1])
                else:
                    retval.add(s)

            return retval

        else:
            blob_iterator = bucket.list_blobs(prefix=key)
            for blob in blob_iterator:
                suffix = re.sub('^' + key, '', blob.name)
                if suffix == '':
                    return json.loads(blob.download_as_string())

                path = suffix.split('/')
                path = [p for p in path if p != '']
                current_dict = retval
                for subdir in path[:-1]:
                    if subdir != '':
                        if subdir not in current_dict.keys():
                            current_dict[subdir] = {}
                        current_dict = current_dict[subdir]

                try:
                    current_dict[path[-1]] = json.loads(
                        blob.download_as_string())
                except BaseException:
                    pass

        if not any(retval):
            return None
        else:
            return retval

    def _delete(self, key):
        self.meta_store._delete_file(key)

    def _set(self, key, value):
        no_retries = 10
        sleep_time = 1
        for i in range(no_retries):
            try:
                self.meta_store._get_bucket_obj().blob(key) \
                    .upload_from_string(json.dumps(value))
                break
            except BaseException as e:
                self.logger.error('uploading data raised an exception:')
                self.logger.exception(e)

            time.sleep(sleep_time)



import sys
import argparse
import logging
import json
import re
import os
import uuid
import importlib
import time
import multiprocessing
import six
import traceback
from contextlib import closing
import numpy as np

from .local_queue import LocalQueue
from .pubsub_queue import PubsubQueue
from .sqs_queue import SQSQueue
from .gcloud_worker import GCloudWorkerManager
from .ec2cloud_worker import EC2WorkerManager
from .hyperparameter import HyperparameterParser
from .util import rand_string, Progbar, rsync_cp
from .experiment import create_experiment

from . import model
from . import auth
from . import git_util
from . import local_worker
from . import fs_tracker


logging.basicConfig()


def main(args=sys.argv):
    logger = logging.getLogger('studio-runner')
    parser = argparse.ArgumentParser(
        description='Studio runner. \
                     Usage: studio run <runner_arguments> \
                     script <script_arguments>')
    parser.add_argument('--config', help='configuration file', default=None)
    parser.add_argument('--project', help='name of the project', default=None)
    parser.add_argument(
        '--experiment', '-e',
        help='Name of the experiment. If none provided, ' +
             'random uuid will be generated',
        default=None)

    parser.add_argument(
        '--guest',
        help='Guest mode (does not require db credentials)',
        action='store_true')

    parser.add_argument(
        '--force-git',
        help='If run in a git directory, force running the experiment ' +
             'even if changes are not commited',
        action='store_true')

    parser.add_argument(
        '--gpus',
        help='Number of gpus needed to run the experiment',
        type=int,
        default=None)

    parser.add_argument(
        '--cpus',
        help='Number of cpus needed to run the experiment' +
             ' (used to configure cloud instance)',
        type=int,
        default=None)

    parser.add_argument(
        '--ram',
        help='Amount of RAM needed to run the experiment' +
             ' (used to configure cloud instance), ex: 10G, 10GB',
        default=None)

    parser.add_argument(
        '--gpuMem',
        help='Amount of GPU RAM needed to run the experiment',
        default=None)

    parser.add_argument(
        '--hdd',
        help='Amount of hard drive space needed to run the experiment' +
             ' (used to configure cloud instance), ex: 10G, 10GB',
        default=None)

    parser.add_argument(
        '--queue', '-q',
        help='Name of the remote execution queue',
        default=None)

    parser.add_argument(
        '--cloud',
        help='Cloud execution mode. Could be gcloud, gcspot, ec2 or ec2spot',
        default=None)

    parser.add_argument(
        '--bid',
        help='Spot instance price bid, specified in USD or in percentage ' +
             'of on-demand instance price. Default is %(default)s',
        default='100%')

    parser.add_argument(
        '--capture-once', '-co',
        help='Name of the immutable artifact to be captured. ' +
        'It will be captured once before the experiment is run',
        default=[], action='append')

    parser.add_argument(
        '--capture', '-c',
        help='Name of the mutable artifact to be captured continuously',
        default=[], action='append')

    parser.add_argument(
        '--reuse', '-r',
        help='Name of the artifact from another experiment to use',
        default=[], action='append')

    parser.add_argument(
        '--verbose', '-v',
        help='Verbosity level. Allowed values: ' +
             'debug, info, warn, error, crit ' +
             'or numerical value of logger levels.',
        default=None)

    parser.add_argument(
        '--metric', '-m',
        help='Metric to show in the summary of the experiment, ' +
             'and to base hyperparameter search on. ' +
             'Refers a scalar value in tensorboard log ' +
             'example: --metric=val_loss[:final | :min | :max] to report ' +
             'validation loss in the end of the keras experiment ' +
             '(or smallest or largest throughout the experiment for :min ' +
             'and :max respectively)',
        default=None)

    parser.add_argument(
        '--hyperparam', '-hp',
        help='Try out multiple values of a certain parameter. ' +
             'For example, --hyperparam=learning_rate:0.01:0.1:l10 ' +
             'will instantiate 10 versions of the script, replace ' +
             'learning_rate with a one of the 10 values for learning ' +
             'rate that lies on a log grid from 0.01 to 0.1, create '
             'experiments and place them in the queue.',
             default=[], action='append')

    parser.add_argument(
        '--num-workers',
        help='Number of local or cloud workers to spin up',
        type=int,
        default=None)

    parser.add_argument(
        '--python-pkg',
        help='Python package not present in the current environment ' +
             'that is needed for experiment. Only compatible with ' +
             'remote and cloud workers for now',
        default=[], action='append')

    parser.add_argument(
        '--ssh-keypair',
        help='Name of the SSH keypair used to access the EC2 ' +
             'instances directly',
        default=None)

    parser.add_argument(
        '--optimizer', '-opt',
        help='Name of optimizer to use, by default is grid search. ' +
        'The name of the optimizer must either be in ' +
        'studio/optimizer_plugins ' +
        'directory or the path to the optimizer source file ' +
        'must be supplied. ',
        default='grid')

    parser.add_argument(
        '--cloud-timeout',
        help="Time (in seconds) that cloud workers wait for messages. " +
             "If negative, " +
             "wait for the first message in the queue indefinitely " +
             "and shut down " +
             "as soon as no new messages are available. " +
             "If zero, don't wait at all." +
             "Default value is %(default)d",
        type=int,
        default=300)

    parser.add_argument(
        '--user-startup-script',
        help='Path of script to run immediately ' +
             'before running the remote worker',
        default=None)

    parser.add_argument(
        '--branch',
        help='Branch of studioml to use when running remote worker, useful ' +
             'for debugging pull requests. Default is current',
        default=None)

    parser.add_argument(
        '--max-duration',
        help='Max experiment runtime (i.e. time after which experiment ' +
             'should be killed no matter what.).  Examples of values ' +
             'might include 5h, 48h2m10s',
        default=None)

    parser.add_argument(
        '--lifetime',
        help='Max experiment lifetime (i.e. wait time after which ' +
             'experiment loses relevance and should not be started)' +
             '  Examples include 240h30m10s',
        default=None)

    parser.add_argument(
        '--container',
        help='Singularity container in which experiment should be run. ' +
             'Assumes that container has all dependencies installed',
        default=None
    )

    # detect which argument is the script filename
    # and attribute all arguments past that index as related to the script
    py_suffix_args = [i for i, arg in enumerate(args) if arg.endswith('.py')]
    rerun = False
    if len(py_suffix_args) < 1:
        print('None of the arugments end with .py')
        (runner_args, other_args) = parser.parse_known_args(args[1:])
        if len(other_args) == 0:
            print("Trying to run a container job")
            assert runner_args.container is not None
            exec_filename = None
        elif len(other_args) == 1:
            print("Treating last argument as experiment key to rerun")
            rerun = True
            experiment_key = args[-1]
        else:
            print("Too many extra arguments - should be either none " +
                  "for container job or one for experiment re-run")
            sys.exit(1)
    else:
        script_index = py_suffix_args[0]
        exec_filename, other_args = args[script_index], args[script_index + 1:]
        runner_args = parser.parse_args(args[1:script_index])

    # TODO: Queue the job based on arguments and only then execute.

    config = model.get_config(runner_args.config)

    if runner_args.verbose:
        config['verbose'] = runner_args.verbose

    if runner_args.guest:
        config['database']['guest'] = True

    if runner_args.container:
        runner_args.capture_once.append(
            runner_args.container + ':_singularity')

    verbose = model.parse_verbosity(config['verbose'])
    logger.setLevel(verbose)

    if git_util.is_git() and not git_util.is_clean() and not rerun:
        logger.warn('Running from dirty git repo')
        if not runner_args.force_git:
            logger.error(
                'Specify --force-git to run experiment from dirty git repo')
            sys.exit(1)

    resources_needed = parse_hardware(runner_args, config['resources_needed'])
    logger.debug('resources requested: ')
    logger.debug(str(resources_needed))

    artifacts = {}
    artifacts.update(parse_artifacts(runner_args.capture, mutable=True))
    artifacts.update(parse_artifacts(runner_args.capture_once, mutable=False))
    with model.get_db_provider(config) as db:
        artifacts.update(parse_external_artifacts(runner_args.reuse, db))

    if runner_args.branch:
        config['cloud']['branch'] = runner_args.branch

    if runner_args.user_startup_script:
        config['cloud']['user_startup_script'] = \
            runner_args.user_startup_script

    if runner_args.lifetime:
        config['experimentLifetime'] = runner_args.lifetime

    if any(runner_args.hyperparam):
        if runner_args.optimizer is "grid":
            experiments = add_hyperparam_experiments(
                exec_filename,
                other_args,
                runner_args,
                artifacts,
                resources_needed,
                logger)

            queue_name = submit_experiments(
                experiments,
                config=config,
                logger=logger,
                queue_name=runner_args.queue,
                cloud=runner_args.cloud)

            spin_up_workers(
                runner_args,
                config,
                resources_needed,
                queue_name=queue_name,
                verbose=verbose)
        else:
            opt_modulepath = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                "optimizer_plugins",
                runner_args.optimizer + ".py")
            if not os.path.exists(opt_modulepath):
                opt_modulepath = os.path.abspath(
                    os.path.expanduser(runner_args.optimizer))
            logger.info('optimizer path: %s' % opt_modulepath)

            assert os.path.exists(opt_modulepath)
            sys.path.append(os.path.dirname(opt_modulepath))
            opt_module = importlib.import_module(
                os.path.basename(opt_modulepath.replace(".py", '')))

            h = HyperparameterParser(runner_args, logger)
            hyperparams = h.parse()
            optimizer = getattr(
                opt_module,
                "Optimizer")(
                hyperparams,
                config['optimizer'],
                logger)

            workers_started = False
            queue_name = runner_args.queue
            while not optimizer.stop():
                hyperparam_pop = optimizer.ask()
                hyperparam_tuples = h.convert_to_tuples(hyperparam_pop)

                experiments = add_hyperparam_experiments(
                    exec_filename,
                    other_args,
                    runner_args,
                    artifacts,
                    resources_needed,
                    logger,
                    optimizer=optimizer,
                    hyperparam_tuples=hyperparam_tuples)

                queue_name = submit_experiments(
                    experiments,
                    config=config,
                    logger=logger,
                    cloud=runner_args.cloud,
                    queue_name=queue_name)

                if not workers_started:
                    spin_up_workers(
                        runner_args,
                        config,
                        resources_needed,
                        queue_name=queue_name,
                        verbose=verbose)
                    workers_started = True

                fitnesses, behaviors = get_experiment_fitnesses(
                    experiments, optimizer, config, logger)

                # for i, hh in enumerate(hyperparam_pop):
                #     print fitnesses[i]
                #     for hhh in hh:
                #         print hhh
                try:
                    optimizer.tell(hyperparam_pop, fitnesses, behaviors)
                except BaseException:
                    optimizer.tell(hyperparam_pop, fitnesses)

                try:
                    optimizer.disp()
                except BaseException:
                    logger.warn('Optimizer has no disp() method')
    else:
        if rerun:
            with model.get_db_provider(config) as db:
                experiment = db.get_experiment(experiment_key)
                new_key = runner_args.experiment if runner_args.experiment \
                    else experiment_key + '_rerun' + str(uuid.uuid4())
                experiment.key = new_key
                for _, art in six.iteritems(experiment.artifacts):
                    art['mutable'] = False

                experiments = [experiment]

        else:
            experiments = [create_experiment(
                filename=exec_filename,
                args=other_args,
                experiment_name=runner_args.experiment,
                project=runner_args.project,
                artifacts=artifacts,
                resources_needed=resources_needed,
                metric=runner_args.metric,
                max_duration=runner_args.max_duration,
            )]

        queue_name = submit_experiments(
            experiments,
            config=config,
            logger=logger,
            cloud=runner_args.cloud,
            queue_name=runner_args.queue)

        spin_up_workers(
            runner_args,
            config,
            resources_needed,
            queue_name=queue_name,
            verbose=verbose)

    return


def add_experiment(args):
    try:
        config, python_pkg, e = args
        e.pythonenv = add_packages(e.pythonenv, python_pkg)
        with model.get_db_provider(config) as db:
            db.add_experiment(e)
    except BaseException:
        traceback.print_exc()
        raise
    return e


def get_worker_manager(config, cloud=None, verbose=10):
    if cloud is None:
        return None

    assert cloud in ['gcloud', 'gcspot', 'ec2', 'ec2spot']
    logger = logging.getLogger('runner.get_worker_manager')
    logger.setLevel(verbose)

    auth_cookie = None if config['database'].get('guest') \
        else os.path.join(
        auth.TOKEN_DIR,
        config['database']['apiKey']
    )

    branch = config['cloud'].get('branch')

    logger.info('using branch {}'.format(branch))

    if cloud in ['gcloud', 'gcspot']:

        cloudconfig = config['cloud']['gcloud']
        worker_manager = GCloudWorkerManager(
            auth_cookie=auth_cookie,
            zone=cloudconfig['zone'],
            branch=branch,
            user_startup_script=config['cloud'].get('user_startup_script')
        )

    if cloud in ['ec2', 'ec2spot']:
        worker_manager = EC2WorkerManager(
            auth_cookie=auth_cookie,
            branch=branch,
            user_startup_script=config['cloud'].get('user_startup_script')
        )
    return worker_manager


def get_queue(queue_name=None, cloud=None, verbose=10):
    if queue_name is None:
        if cloud in ['gcloud', 'gcspot']:
            queue_name = 'pubsub_' + str(uuid.uuid4())
        elif cloud in ['ec2', 'ec2spot']:
            queue_name = 'sqs_' + str(uuid.uuid4())
        else:
            queue_name = 'local'

    if queue_name.startswith('ec2') or \
       queue_name.startswith('sqs'):
        return SQSQueue(queue_name, verbose=verbose)
    elif queue_name == 'local':
        return LocalQueue(verbose=verbose)
    else:
        return PubsubQueue(queue_name, verbose=verbose)


def spin_up_workers(
        runner_args,
        config,
        resources_needed,
        queue_name,
        verbose=10):
    cloud = runner_args.cloud
    if cloud:

        assert cloud in ['gcloud', 'gcspot', 'ec2', 'ec2spot']
        worker_manager = get_worker_manager(config, cloud, verbose=verbose)

        if cloud == 'gcloud' or \
           cloud == 'ec2':

            if runner_args.num_workers:
                num_workers = int(runner_args.num_workers)
            else:
                num_workers = 1

            for i in range(num_workers):
                worker_manager.start_worker(
                    queue_name, resources_needed,
                    ssh_keypair=runner_args.ssh_keypair,
                    timeout=runner_args.cloud_timeout)
        else:
            assert runner_args.bid is not None
            if runner_args.num_workers:
                start_workers = runner_args.num_workers
                queue_upscaling = False
            else:
                start_workers = 1
                queue_upscaling = True

            worker_manager.start_spot_workers(
                queue_name,
                runner_args.bid,
                resources_needed,
                start_workers=start_workers,
                queue_upscaling=queue_upscaling,
                ssh_keypair=runner_args.ssh_keypair,
                timeout=runner_args.cloud_timeout)

    elif queue_name == 'local':
        worker_args = ['studio-local-worker']

        if runner_args.config:
            worker_args += ['--config=' + runner_args.config]

        if runner_args.guest:
            worker_args += ['--guest']

        # logger.info('worker args: {}'.format(worker_args))
        if not runner_args.num_workers or int(runner_args.num_workers) == 1:
            if 'STUDIOML_DUMMY_MODE' not in os.environ:
                local_worker.main(worker_args)
        else:
            raise NotImplementedError("Multiple local workers are not " +
                                      "implemented yet")


def submit_experiments(
        experiments,
        config,
        logger,
        cloud=None,
        queue_name=None,
        python_pkg=[]):

    num_experiments = len(experiments)
    verbose = model.parse_verbosity(config['verbose'])

    start_time = time.time()
    n_workers = min(multiprocessing.cpu_count() * 2, num_experiments)

    with closing(multiprocessing.Pool(n_workers, maxtasksperchild=20)) as p:
        experiments = p.imap_unordered(add_experiment,
                                       zip([config] * num_experiments,
                                           [python_pkg] *
                                           num_experiments,
                                           experiments),
                                       chunksize=1)
        p.close()
        p.join()

    '''
    experiments = [add_experiment(e) for e in
                    zip([config] * num_experiments,
                        [python_pkg] *
                        num_experiments,
                        experiments)]
    '''

    logger.info("Added %s experiments in %s seconds" %
                (num_experiments, int(time.time() - start_time)))

    queue = get_queue(queue_name, cloud, verbose)
    for e in experiments:
        queue.enqueue(json.dumps({
            'experiment': e.__dict__,
            'config': config}))

    return queue.get_name()


def get_experiment_fitnesses(experiments, optimizer, config, logger):
    with model.get_db_provider() as db:
        progbar = Progbar(len(experiments), interval=0.0)
        logger.info("Waiting for fitnesses from %s experiments" %
                    len(experiments))

        bad_line_dicts = [dict() for x in range(len(experiments))]
        has_result = [False for i in range(len(experiments))]
        fitnesses = [0.0 for i in range(len(experiments))]
        behaviors = [None for i in range(len(experiments))]
        term_criterion = config['optimizer']['termination_criterion']
        skip_gen_thres = term_criterion['skip_gen_thres']
        skip_gen_timeout = term_criterion['skip_gen_timeout']
        result_timestamp = time.time()

        while sum(has_result) < len(experiments):
            for i, experiment in enumerate(experiments):
                if float(sum(has_result)) / len(experiments) >= skip_gen_thres\
                        and time.time() - result_timestamp > skip_gen_timeout:
                    logger.warn(
                        "Skipping to next gen with %s of solutions evaled" %
                        (float(
                            sum(has_result)) /
                            len(experiments)))
                    has_result = [True] * len(experiments)
                    break
                if has_result[i]:
                    continue
                returned_experiment = db.get_experiment(experiment.key,
                                                        getinfo=True)
                # try:
                #     experiment_output = returned_experiment.info['logtail']
                # except BaseException:
                #     logger.warn('Cannot access "logtail" in experiment.info')
                output = db._get_experiment_logtail(
                    returned_experiment)
                if output is None:
                    continue

                for j, line in enumerate(output):

                    if line.startswith(
                            "Traceback (most recent call last):") and \
                            j not in bad_line_dicts[i]:
                        logger.warn("Experiment %s: error"
                                    " discovered in output" %
                                    returned_experiment.key)
                        logger.warn("".join(output[j:]))
                        bad_line_dicts[i][j] = True

                    if line.startswith("Behavior") or \
                            line.startswith("behavior"):
                        try:
                            behavior = eval(line.rstrip().split(':')[1])
                            if isinstance(behavior, np.ndarray):
                                pass
                            elif isinstance(behavior, list):
                                behavior = np.array(behavior)
                            else:
                                raise

                        except BaseException:
                            if j not in bad_line_dicts[i]:
                                logger.warn(
                                    'Experiment %s: error parsing or invalid'
                                    ' behavior' %
                                    returned_experiment.key)
                                logger.warn(line)
                                bad_line_dicts[i][j] = True
                        else:
                            behaviors[i] = behavior

                    if line.startswith("Fitness") or \
                            line.startswith("fitness"):
                        try:
                            fitness = float(line.rstrip().split(':')[1])
                            # assert fitness >= 0.0
                        except BaseException:
                            if j not in bad_line_dicts[i]:
                                logger.warn(
                                    'Experiment %s: error parsing or invalid'
                                    ' fitness' %
                                    returned_experiment.key)
                                logger.warn(line)
                                bad_line_dicts[i][j] = True
                        else:
                            if fitness < 0.0:
                                logger.warn('Experiment %s: returned'
                                            ' fitness is less than zero,'
                                            ' setting it to zero' %
                                            returned_experiment.key)
                                fitness = 0.0

                            fitnesses[i] = fitness
                            has_result[i] = True
                            progbar.add(1)
                            result_timestamp = time.time()
                            break

            time.sleep(config['sleep_time'])
        print
        return fitnesses, behaviors


def parse_artifacts(art_list, mutable):
    retval = {}
    url_schema = re.compile('^https{0,1}://')
    s3_schema = re.compile('s3://')
    gcs_schema = re.compile('gs://')
    dhub_schema = re.compile('dockerhub://')
    shub_schema = re.compile('shub://')

    for entry in art_list:
        path = re.sub(':[^:]*\Z', '', entry)
        tag = re.sub('.*:(?=[^:]*\Z)', '', entry)

        if url_schema.match(entry):
            assert not mutable, \
                'artifacts specfied by url can only be immutable'
            retval[tag] = {
                'url': path,
                'mutable': False
            }
        elif s3_schema.match(entry) or \
                gcs_schema.match(entry) or \
                dhub_schema.match(entry) or \
                shub_schema.match(entry):

            assert not mutable, \
                'artifacts specfied by url can only be immutable'
            retval[tag] = {
                'qualified': path,
                'mutable': False
            }
        else:
            retval[tag] = {
                'local': os.path.expanduser(path),
                'mutable': mutable
            }
            if not mutable:
                assert os.path.exists(retval[tag]['local'])
    return retval


def parse_external_artifacts(art_list, db):
    retval = {}
    for entry in art_list:
        external = re.sub(':.*', '', entry)
        tag = re.sub('.*:', '', entry)

        experiment_key = re.sub('/.*', '', external)
        external_tag = re.sub('.*/', '', external)
        experiment = db.get_experiment(experiment_key, getinfo=False)

        retval[tag] = {
            'key': experiment.artifacts[external_tag]['key'],
            'mutable': False
        }
    return retval


def parse_hardware(runner_args, config={}):
    resources_needed = {}
    parse_list = ['gpus', 'cpus', 'ram', 'hdd', 'gpuMem']
    for key in parse_list:
        from_args = runner_args.__dict__.get(key)
        from_config = config.get(key)
        if from_args is not None:
            resources_needed[key] = from_args
        elif from_config is not None:
            resources_needed[key] = from_config

    return resources_needed


def add_hyperparam_experiments(
        exec_filename,
        other_args,
        runner_args,
        artifacts,
        resources_needed,
        logger,
        optimizer=None,
        hyperparam_tuples=None):

    experiment_name_base = runner_args.experiment if runner_args.experiment \
        else str(uuid.uuid4())

    project = runner_args.project if runner_args.project else \
        ('hyperparam_' + experiment_name_base)

    workspace_orig = artifacts['workspace']['local'] \
        if 'workspace' in artifacts.keys() else '.'

    ignore_arg = ''
    ignore_filepath = os.path.join(workspace_orig, ".studioml_ignore")
    if os.path.exists(ignore_filepath) and \
            not os.path.isdir(ignore_filepath):
        ignore_arg = "--exclude-from=%s" % ignore_filepath

    def create_experiments(hyperparam_tuples):
        experiments = []
        # experiment_names = {}
        for hyperparam_tuple in hyperparam_tuples:
            experiment_name = experiment_name_base
            experiment_name += "__opt__%s__%s" % (rand_string(32),
                                                  int(time.time()))
            experiment_name = experiment_name.replace('.', '_')

            workspace_new = fs_tracker.get_artifact_cache(
                'workspace', experiment_name)

            current_artifacts = artifacts.copy()
            current_artifacts.update({
                'workspace': {
                    'local': workspace_new,
                    'mutable': True
                }
            })

            rsync_cp(workspace_orig, workspace_new, ignore_arg, logger)
            # shutil.copytree(workspace_orig, workspace_new)

            for param_name, param_value in six.iteritems(hyperparam_tuple):
                if isinstance(param_value, np.ndarray):
                    array_filepath = '/tmp/%s.npy' % rand_string(32)
                    np.save(array_filepath, param_value)
                    assert param_name not in current_artifacts
                    current_artifacts[param_name] = {'local': array_filepath,
                                                     'mutable': False}
                else:
                    with open(os.path.join(workspace_new, exec_filename),
                              'r') as f:
                        script_text = f.read()

                    script_text = re.sub(
                        '\\b' +
                        param_name +
                        '\\b(?=[^=]*\\n)',
                        str(param_value),
                        script_text)

                    with open(os.path.join(workspace_new, exec_filename),
                              'w') as f:
                        f.write(script_text)

            experiments.append(create_experiment(
                filename=exec_filename,
                args=other_args,
                experiment_name=experiment_name,
                project=project,
                artifacts=current_artifacts,
                resources_needed=resources_needed,
                metric=runner_args.metric,
                max_duration=runner_args.max_duration,
            ))
        return experiments

    if optimizer is not None:
        experiments = create_experiments(hyperparam_tuples)
    else:
        h = HyperparameterParser(runner_args, logger)
        hyperparam_tuples = h.convert_to_tuples(h.parse())
        experiments = create_experiments(hyperparam_tuples)

    return experiments


def add_packages(list1, list2):
    pkg_dict = {re.sub('==.+', '', pkg): pkg for pkg in list1 + list2}
    return [pkg for _, pkg in six.iteritems(pkg_dict)]


if __name__ == "__main__":
    main()



import hashlib
from io import StringIO
from datetime import timedelta
import re
import random
import string
import struct
import time
import sys
import shutil
import subprocess
import os
import numpy as np
import requests
import six

from tensorflow.core.util import event_pb2

import boto3

DAY = 86400
HOUR = 3600
MINUTE = 60


def remove_backspaces(line):
    splitline = re.split('(\x08+)', line)
    try:
        splitline = [unicode(s, 'utf-8') for s in splitline]
    except NameError:
        splitline = [str(s) for s in splitline]

    buf = StringIO()
    for i in range(0, len(splitline) - 1, 2):
        buf.write(splitline[i][:-len(splitline[i + 1])])

    if len(splitline) % 2 == 1:
        buf.write(splitline[-1])

    return buf.getvalue()


def sha256_checksum(filename, block_size=65536):
    sha256 = hashlib.sha256()
    with open(filename, 'rb') as f:
        for block in iter(lambda: f.read(block_size), b''):
            sha256.update(block)
    return sha256.hexdigest()


def rand_string(length):
    return "".join([random.choice(string.ascii_letters + string.digits)
                    for n in range(length)])


def event_reader(fileobj):

    if isinstance(fileobj, str):
        fileobj = open(fileobj, 'rb')

    header_len = 12
    footer_len = 4
    size_len = 8

    while True:
        try:
            data_len = struct.unpack('Q', fileobj.read(size_len))[0]
            fileobj.read(header_len - size_len)

            data = fileobj.read(data_len)

            event = None
            event = event_pb2.Event()
            event.ParseFromString(data)

            fileobj.read(footer_len)
            yield event
        except BaseException:
            break

    fileobj.close()


def rsync_cp(source, dest, ignore_arg='', logger=None):
    if os.path.exists(dest):
        shutil.rmtree(dest) if os.path.isdir(dest) else os.remove(dest)
    os.makedirs(dest)

    if ignore_arg != '':
        source += "/"
        tool = 'rsync'
        args = [tool, ignore_arg, '-aHAXE', source, dest]
    else:
        os.rmdir(dest)
        tool = 'cp'
        args = [tool, '-pR', source, dest]

    pcp = subprocess.Popen(args, stdout=subprocess.PIPE,
                           stderr=subprocess.STDOUT)
    cpout, _ = pcp.communicate()
    if pcp.returncode != 0 and logger is not None:
        logger.info('%s returned non-zero exit code. Output:' % tool)
        logger.info(cpout)


class Progbar(object):
    """Displays a progress bar.

    # Arguments
        target: Total number of steps expected, None if unknown.
        interval: Minimum visual progress update interval (in seconds).
    """

    def __init__(self, target, width=30, verbose=1, interval=0.05):
        self.width = width
        if target is None:
            target = -1
        self.target = target
        self.sum_values = {}
        self.unique_values = []
        self.start = time.time()
        self.last_update = 0
        self.interval = interval
        self.total_width = 0
        self.seen_so_far = 0
        self.verbose = verbose

    def update(self, current, values=None, force=False):
        """Updates the progress bar.

        # Arguments
            current: Index of current step.
            values: List of tuples (name, value_for_last_step).
                The progress bar will display averages for these values.
            force: Whether to force visual progress update.
        """
        values = values or []
        for k, v in values:
            if k not in self.sum_values:
                self.sum_values[k] = [v * (current - self.seen_so_far),
                                      current - self.seen_so_far]
                self.unique_values.append(k)
            else:
                self.sum_values[k][0] += v * (current - self.seen_so_far)
                self.sum_values[k][1] += (current - self.seen_so_far)
        self.seen_so_far = current

        now = time.time()
        if self.verbose == 1:
            if not force and (now - self.last_update) < self.interval:
                return

            prev_total_width = self.total_width
            sys.stdout.write('\b' * prev_total_width)
            sys.stdout.write('\r')

            if self.target is not -1:
                numdigits = int(np.floor(np.log10(self.target))) + 1
                barstr = '%%%dd/%%%dd [' % (numdigits, numdigits)
                bar = barstr % (current, self.target)
                prog = float(current) / self.target
                prog_width = int(self.width * prog)
                if prog_width > 0:
                    bar += ('=' * (prog_width - 1))
                    if current < self.target:
                        bar += '>'
                    else:
                        bar += '='
                bar += ('.' * (self.width - prog_width))
                bar += ']'
                sys.stdout.write(bar)
                self.total_width = len(bar)

            if current:
                time_per_unit = (now - self.start) / current
            else:
                time_per_unit = 0
            eta = time_per_unit * (self.target - current)
            info = ''
            if current < self.target and self.target is not -1:
                info += ' - ETA: %ds' % eta
            else:
                info += ' - %ds' % (now - self.start)
            for k in self.unique_values:
                info += ' - %s:' % k
                if isinstance(self.sum_values[k], list):
                    avg = np.mean(
                        self.sum_values[k][0] / max(1, self.sum_values[k][1]))
                    if abs(avg) > 1e-3:
                        info += ' %.4f' % avg
                    else:
                        info += ' %.4e' % avg
                else:
                    info += ' %s' % self.sum_values[k]

            self.total_width += len(info)
            if prev_total_width > self.total_width:
                info += ((prev_total_width - self.total_width) * ' ')

            sys.stdout.write(info)
            sys.stdout.flush()

            if current >= self.target:
                sys.stdout.write('\n')

        if self.verbose == 2:
            if current >= self.target:
                info = '%ds' % (now - self.start)
                for k in self.unique_values:
                    info += ' - %s:' % k
                    avg = np.mean(
                        self.sum_values[k][0] / max(1, self.sum_values[k][1]))
                    if avg > 1e-3:
                        info += ' %.4f' % avg
                    else:
                        info += ' %.4e' % avg
                sys.stdout.write(info + "\n")

        self.last_update = now

    def add(self, n, values=None):
        self.update(self.seen_so_far + n, values)


def download_file(url, local_path, logger=None):
    response = requests.get(
        url,
        stream=True)
    if logger:
        logger.info(("Trying to download file at url {} to " +
                     "local path {}").format(url, local_path))

    if response.status_code == 200:
        with open(local_path, 'wb') as f:
            for chunk in response:
                f.write(chunk)
    elif logger:
        logger.info("Response error with code {}"
                    .format(response.status_code))

    return response


def upload_file(url, local_path, logger=None):
    if logger:
        logger.info(("Trying to upload file {} to " +
                     "url {}").format(local_path, url))
    tic = time.time()
    with open(local_path, 'rb') as f:
        resp = requests.put(url, data=f)

    if resp.status_code != 200 and logger:
        logger.error(str(resp.reason))

    if logger:
        logger.debug('File upload done in {} s'
                     .format(time.time() - tic))


def download_file_from_qualified(qualified, local_path, logger=None):
    if qualified.startswith('dockerhub://') or \
       qualified.startswith('shub://'):
        return

    assert qualified.startswith('s3://') or \
        qualified.startswith('gs://')

    bucket = qualified.split('/')[2]
    key = '/'.join(qualified.split('/')[3:])

    if logger is not None:
        logger.debug(('Downloading file from bucket {} ' +
                      ' and key {} to local path {}')
                     .format(bucket, key, local_path))

    if qualified.startswith('s3://'):
        boto3.client('s3').download_file(bucket, key, local_path)
    else:
        raise NotImplementedError


def has_aws_credentials():
    return boto3.client('s3')._request_signer._credentials is not None


def retry(f,
          no_retries=5, sleep_time=1,
          exception_class=BaseException, logger=None):
    for i in range(no_retries):
        try:
            return f()
        except exception_class as e:
            if logger:
                logger.info(
                    ('Exception {} is caught, ' +
                     'sleeping {}s and retrying (attempt {} of {})')
                    .format(e, sleep_time, i, no_retries))
            time.sleep(sleep_time)


def compression_to_extension(compression):
    return _compression_to_extension_taropt(compression)[0]


def compression_to_taropt(compression):
    return _compression_to_extension_taropt(compression)[1]


def _compression_to_extension_taropt(compression):
    default_compression = 'none'
    if compression is None:
        compression = default_compression

    compression = compression.lower()

    if compression == 'bzip2':
        return '.bz2', '--bzip2'

    elif compression == 'gzip':
        return '.gz', '--gzip'

    elif compression == 'xz':
        return '.xz', '--xz'

    elif compression == 'lzma':
        return '.lzma', '--lzma'

    elif compression == 'lzop':
        return '.lzop', '--lzop'

    elif compression == 'none':
        return '', ''

    raise ValueError('Unknown compression method {}'
                     .format(compression))


def timeit(method):

    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()

        line = '%r (%r, %r) %2.2f sec' % \
            (method.__name__, args, kw, te - ts)

        try:
            logger = args[0].logger
            logger.info(line)
        except BaseException:
            print(line)

        return result

    return timed


def sixdecode(s):
    if isinstance(s, six.string_types):
        return s
    if isinstance(s, six.binary_type):
        return s.decode('utf8')

    raise TypeError("Unknown type of " + str(s))


duration_regex = re.compile(
    r'((?P<hours>-?\d+?)h)?((?P<minutes>-?\d+?)m)?((?P<seconds>-?\d+?)s)?')

# parse_duration parses strings into time delta values that python can
# deal with.  Examples include 12h, 11h60m, 719m60s, 11h3600s
#


def parse_duration(duration_str):
    parts = duration_regex.match(duration_str)
    if not parts:
        return
    parts = parts.groupdict()
    time_params = {}
    for (name, param) in six.iteritems(parts):
        if param:
            time_params[name] = int(param)
    retval = timedelta(**time_params)
    return retval


def str2duration(s):
    return parse_duration(s.lower())



import os
import getpass
import time
import json
import shutil
import atexit
import tempfile
import logging

try:
    from apscheduler.schedulers.background import BackgroundScheduler
except BaseException:
    BackgroundScheduler = None

from .util import rand_string

logging.basicConfig()

TOKEN_DIR = os.path.expanduser('~/.studioml/keys')
HOUR = 3600
HALF_HOUR = 1800
API_KEY_COOLDOWN = 900
SLEEP_TIME = 0.5
MAX_NUM_RETRIES = 100


_auth_singleton = None


def get_auth(
        firebase,
        use_email_auth=False,
        email=None,
        password=None,
        blocking=True):

    global _auth_singleton
    if _auth_singleton is None:
        _auth_singleton = FirebaseAuth(
            firebase,
            use_email_auth,
            email,
            password,
            blocking)

    return _auth_singleton


class FirebaseAuth(object):
    def __init__(
            self,
            firebase,
            use_email_auth=False,
            email=None,
            password=None,
            blocking=True):
        if not os.path.exists(TOKEN_DIR):
            os.makedirs(TOKEN_DIR)

        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.DEBUG)

        self.firebase = firebase
        self.user = {}
        self.use_email_auth = use_email_auth
        if use_email_auth:
            if email and password:
                self.email = email
                self.password = password
            else:
                self.email = input(
                    'Firebase token is not found or expired! ' +
                    'You need to re-login. (Or re-run with ' +
                    'studio/studio-runner ' +
                    'with --guest option ) '
                    '\nemail:')
                self.password = getpass.getpass('password:')

        self.expired = True
        self._update_user()

        if self.expired and blocking:
            print('Authentication required! Either specify ' +
                  'use_email_auth in config file, or run '
                  'studio and go to webui ' +
                  '(localhost:5000 by default) '
                  'to authenticate using google credentials')
            while self.expired:
                time.sleep(1)
                self._update_user()

        self.sched = BackgroundScheduler()
        self.sched.start()
        self.sched.add_job(self._update_user, 'interval', minutes=31)
        atexit.register(self.sched.shutdown)

    def _update_user(self):
        api_key = os.path.join(TOKEN_DIR, self.firebase.api_key)
        if not os.path.exists(api_key):
                # refresh tokens don't expire, hence we can
                # use them forever once obtained
                # or (time.time() - os.path.getmtime(api_key)) > HOUR:
            if self.use_email_auth:
                self.sign_in_with_email()
                self.expired = False
            else:
                self.expired = True
        else:
            # If json file fails to load, try again
            counter = 0
            user = None
            while True:
                if user is not None or counter >= MAX_NUM_RETRIES:
                    break
                try:
                    with open(api_key) as f:
                        user = json.loads(f.read())
                except BaseException as e:
                    self.logger.info(e)
                    time.sleep(SLEEP_TIME)
                    counter += 1
            if user is None:
                return

            self.user = user
            if time.time() > self.user.get('expiration', 0):
                counter = 0
                while counter < MAX_NUM_RETRIES:
                    try:
                        self.refresh_token(user['email'], user['refreshToken'])
                        break
                    except BaseException as e:
                        self.logger.info(e)
                        time.sleep(SLEEP_TIME)
                        counter += 1
            else:
                self.expired = False

    def sign_in_with_email(self):
        self.user = \
            self.firebase.auth().sign_in_with_email_and_password(
                self.email,
                self.password)

        # TODO check if credentials worked

        self.user['email'] = self.email

    def refresh_token(self, email, refresh_token):
        api_key = os.path.join(TOKEN_DIR, self.firebase.api_key)
        self.user = self.firebase.auth().refresh(refresh_token)
        self.user['email'] = email
        self.user['expiration'] = time.time() + API_KEY_COOLDOWN
        self.expired = False

        # if not os.path.exists(api_key) or \
        #   time.time() - os.path.getmtime(api_key) > HALF_HOUR:
        # Rename to ensure atomic writes to json file
        # (technically more safe, but slower)

        tmp_api_key = os.path.join(tempfile.gettempdir(),
                                   "api_key_%s" % rand_string(32))
        with open(tmp_api_key, 'w') as f:
            f.write(json.dumps(self.user))
            f.flush()
            os.fsync(f.fileno())
            f.close()
        os.rename(tmp_api_key, api_key)

    def get_token(self):
        if self.expired:
            return None
        return self.user['idToken']

    def get_user_id(self):
        if self.expired:
            return None

        if 'localId' in self.user.keys():
            return self.user['localId']

        return self.user['userId']

    def get_user_email(self):
        if self.expired:
            return None
        # we could also use the get_account_info
        # print self.firebase.auth().get_account_info(self.get_token())
        return self.user['email']


def remove_all_keys():
    keypath = os.path.join(os.path.expanduser('~'), '.studioml', 'keys')
    if os.path.exists(keypath):
        try:
            shutil.rmtree(keypath)
        except OSError:
            pass



import os
import glob
import uuid
import sys
import time

try:
    import pip
except ImportError:
    pip = None

from . import fs_tracker


class Experiment(object):
    """Experiment information."""

    def __init__(self, key, filename, args, pythonenv,
                 project=None,
                 artifacts=None,
                 status='waiting',
                 resources_needed=None,
                 time_added=None,
                 time_started=None,
                 time_last_checkpoint=None,
                 time_finished=None,
                 info={},
                 git=None,
                 metric=None,
                 pythonver=None,
                 max_duration=None):

        self.key = key
        self.filename = filename
        self.args = args if args else []
        self.pythonenv = pythonenv
        self.project = project
        self.pythonver = pythonver if pythonver else sys.version_info[0]

        workspace_path = os.path.abspath('.')
        try:
            model_dir = fs_tracker.get_model_directory(key)
        except BaseException:
            model_dir = None

        self.artifacts = {
            'workspace': {
                'local': workspace_path,
                'mutable': True
            },
            'modeldir': {
                'local': model_dir,
                'mutable': True
            },
            'output': {
                'local': fs_tracker.get_artifact_cache('output', key),
                'mutable': True
            },
            'tb': {
                'local': fs_tracker.get_tensorboard_dir(key),
                'mutable': True
            },
            '_metrics': {
                'local': fs_tracker.get_artifact_cache('_metrics', key),
                'mutable': True
            }
        }
        if artifacts is not None:
            self.artifacts.update(artifacts)

        self.resources_needed = resources_needed
        self.status = status
        self.time_added = time_added
        self.time_started = time_started
        self.time_last_checkpoint = time_last_checkpoint
        self.time_finished = time_finished
        self.info = info
        self.git = git
        self.metric = metric
        self.max_duration = max_duration

    def get_model(self, db):
        modeldir = db.get_artifact(self.artifacts['modeldir'])
        hdf5_files = [
            (p, os.path.getmtime(p))
            for p in
            glob.glob(modeldir + '/*.hdf*') +
            glob.glob(modeldir + '/*.h5')]
        if any(hdf5_files):
            # experiment type - keras
            import keras
            last_checkpoint = max(hdf5_files, key=lambda t: t[1])[0]
            return keras.models.load_model(last_checkpoint)

        if self.info.get('type') == 'tensorflow':
            raise NotImplementedError

        raise ValueError("Experiment type is unknown!")


def create_experiment(
        filename,
        args,
        experiment_name=None,
        project=None,
        artifacts={},
        resources_needed=None,
        metric=None,
        max_duration=None):
    key = experiment_name if experiment_name else \
        str(int(time.time())) + "_" + str(uuid.uuid4())

    packages = []

    for i, pkg in enumerate(
            pip.pip.get_installed_distributions(local_only=True)):
        if pkg._key == 'studioml':
            continue

        if resources_needed is not None and \
           int(resources_needed.get('gpus')) > 0 and \
           (pkg._key == 'tensorflow' or pkg._key == 'tf-nightly'):
            packages.insert(0, pkg._key + '-gpu==' + pkg._version)
            packages.insert(0, pkg._key + '==' + pkg._version)

        elif pkg._key == 'tensorflow-gpu' or pkg._key == 'tf-nightly-gpu':
            packages.insert(0, pkg._key[:-4] + '==' + pkg._version)
        else:
            packages.append(pkg._key + '==' + pkg._version)

    return Experiment(
        key=key,
        filename=filename,
        args=args,
        pythonenv=[p for p in packages],
        project=project,
        artifacts=artifacts,
        resources_needed=resources_needed,
        metric=metric,
        max_duration=max_duration)


def experiment_from_dict(data, info={}):
    return Experiment(
        key=data['key'],
        filename=data['filename'],
        args=data.get('args'),
        pythonenv=data['pythonenv'],
        project=data.get('project'),
        status=data['status'],
        artifacts=data.get('artifacts'),
        resources_needed=data.get('resources_needed'),
        time_added=data['time_added'],
        time_started=data.get('time_started'),
        time_last_checkpoint=data.get('time_last_checkpoint'),
        time_finished=data.get('time_finished'),
        info=info if any(info) else data.get('info'),
        git=data.get('git'),
        metric=data.get('metric'),
        pythonver=data.get('pythonver'),
        max_duration=data.get('max_duration')
    )



import requests
import json
import six
import time
import re

from . import pyrebase
from .auth import get_auth
from .http_artifact_store import HTTPArtifactStore
from .experiment import experiment_from_dict
from .util import retry

import logging
logging.basicConfig()


class HTTPProvider(object):
    """Data provider communicating with API server."""

    def __init__(
            self,
            config,
            verbose=10,
            blocking_auth=True,
            compression=None):
        # TODO: implement connection
        self.url = config.get('serverUrl')
        self.verbose = verbose
        self.logger = logging.getLogger('HTTPProvider')
        self.logger.setLevel(self.verbose)

        self.auth = None
        self.app = pyrebase.initialize_app(config)
        guest = config.get('guest')
        if not guest and 'serviceAccount' not in config.keys():
            self.auth = get_auth(self.app,
                                 config.get("use_email_auth"),
                                 config.get("email"),
                                 config.get("password"),
                                 blocking_auth)

        self.compression = compression
        if self.compression is None:
            self.compression = config.get('compression')

    def add_experiment(self, experiment, userid=None,
                       compression=None):

        headers = self._get_headers()
        compression = compression if compression else self.compression

        for tag, art in six.iteritems(experiment.artifacts):
            if not art['mutable'] and art.get('local') is not None:
                art['hash'] = HTTPArtifactStore(None, None,
                                                compression,
                                                self.verbose) \
                    .get_artifact_hash(art)

        data = {}
        data['experiment'] = experiment.__dict__
        data['compression'] = compression

        def post_request():
            request = requests.post(
                self.url + '/api/add_experiment',
                headers=headers,
                data=json.dumps(data))

            self._raise_detailed_error(request)
            return request

        request = retry(post_request, sleep_time=10, logger=self.logger)
        artifacts = request.json()['artifacts']

        self._update_artifacts(experiment, artifacts)

    def _update_artifacts(self, experiment, artifacts):
        self.logger.debug(str(experiment.artifacts.keys()))
        self.logger.debug(str(artifacts.keys()))

        for tag, art in six.iteritems(experiment.artifacts):
            target_art = artifacts.get(tag)
            if target_art is not None:
                art['key'] = target_art['key']
                art['qualified'] = target_art['qualified']
                art['bucket'] = target_art['bucket']

                if art.get('local'):
                    HTTPArtifactStore(target_art['url'],
                                      target_art['timestamp'],
                                      compression=self.compression,
                                      verbose=self.verbose) \
                        .put_artifact(art)

    def delete_experiment(self, experiment):
        if isinstance(experiment, six.string_types):
            key = experiment
        else:
            key = experiment.key

        headers = self._get_headers()

        def post_request():
            request = requests.post(self.url + '/api/delete_experiment',
                                    headers=headers,
                                    data=json.dumps({"key": key})
                                    )
            self._raise_detailed_error(request)

        post_request()
        # retry(post_request, sleep_time=10, logger=self.logger)

    def get_experiment(self, experiment, getinfo='True'):
        if isinstance(experiment, six.string_types):
            key = experiment
        else:
            key = experiment.key

        headers = self._get_headers()
        try:
            request = requests.post(self.url + '/api/get_experiment',
                                    headers=headers,
                                    data=json.dumps({"key": key})
                                    )

            self._raise_detailed_error(request)
            return experiment_from_dict(request.json()['experiment'])
        except BaseException as e:
            self.logger.info(e)
            return None

    def start_experiment(self, experiment):
        self.checkpoint_experiment(experiment)
        if isinstance(experiment, six.string_types):
            key = experiment
        else:
            key = experiment.key

        headers = self._get_headers()

        def post_request():
            request = requests.post(self.url + '/api/start_experiment',
                                    headers=headers,
                                    data=json.dumps({"key": key})
                                    )
            self._raise_detailed_error(request)
        retry(post_request, sleep_time=10, logger=self.logger)

        if not isinstance(experiment, six.string_types):
            experiment.time_started = time.time()

    def stop_experiment(self, experiment):
        if isinstance(experiment, six.string_types):
            key = experiment
        else:
            key = experiment.key

        headers = self._get_headers()

        def post_request():
            request = requests.post(self.url + '/api/stop_experiment',
                                    headers=headers,
                                    data=json.dumps({"key": key})
                                    )
            self._raise_detailed_error(request)

        retry(post_request, sleep_time=10, logger=self.logger)

    def finish_experiment(self, experiment):
        if isinstance(experiment, six.string_types):
            key = experiment
        else:
            key = experiment.key

        headers = self._get_headers()

        def post_request():
            request = requests.post(self.url + '/api/finish_experiment',
                                    headers=headers,
                                    data=json.dumps({"key": key})
                                    )
            self._raise_detailed_error(request)

        retry(post_request, sleep_time=10, logger=self.logger)
        if not isinstance(experiment, six.string_types):
            experiment.time_finished = time.time()

    def get_user_experiments(self, user=None, blocking=True):
        headers = self._get_headers()
        user = user if user else self._get_userid()

        response = requests.post(
            self.url + '/api/get_user_experiments',
            headers=headers,
            data=json.dumps({"user": user}))

        self._raise_detailed_error(response)
        data = response.json()['experiments']

        experiments = data

        return experiments

    def get_projects(self):
        headers = self._get_headers()
        response = requests.post(
            self.url + '/api/get_projects',
            headers=headers)

        self._raise_detailed_error(response)
        projects = response.json()['projects']

        return projects

    def get_project_experiments(self, project):
        headers = self._get_headers()
        response = requests.post(
            self.url + '/api/get_project_experiments',
            headers=headers,
            data=json.dumps({"project": project}))

        self._raise_detailed_error(response)
        data = response.json()['experiments']

        experiments = data
        return experiments

    def get_artifacts(self, key):
        return {t: a['url'] for t, a in
                six.iteritems(self.get_experiment(key).artifacts)}

    def get_artifact(self, artifact,
                     local_path=None, only_newer='True'):

        if isinstance(artifact, six.string_types):
            experiment_key = re.match(r'.*(?=/)', artifact).group(0)
            artifact_tag = re.search(r'(?<=/)[^/]*\Z', artifact).group(0)
            experiment = self.get_experiment(experiment_key)
            artifact = experiment.artifacts[artifact_tag]

        return HTTPArtifactStore(
            artifact.get('url'),
            timestamp=time.time(),
            verbose=self.verbose
        ).get_artifact(artifact, local_path=local_path)

    def get_users(self):
        headers = self._get_headers()
        response = requests.post(
            self.url + '/api/get_users',
            headers=headers)

        self._raise_detailed_error(response)
        users = response.json()['users']

        return users

    def checkpoint_experiment(self, experiment):
        if isinstance(experiment, six.string_types):
            key = experiment
            experiment = self.get_experiment(key)
        else:
            key = experiment.key

        headers = self._get_headers()

        def post_request():
            request = requests.post(self.url + '/api/checkpoint_experiment',
                                    headers=headers,
                                    data=json.dumps({"key": key}))

            self._raise_detailed_error(request)
            artifacts = request.json()['artifacts']
            return artifacts

        artifacts = retry(post_request, sleep_time=10, logger=self.logger)

        self._update_artifacts(experiment, artifacts)
        experiment.time_last_checkpoint = time.time()

    def refresh_auth_token(self, email, refresh_token):
        if self.auth:
            self.auth.refresh_token(email, refresh_token)

    def register_user(self, userid, email):
        pass

    def _get_headers(self):
        headers = {"content-type": "application/json"}
        if self.auth:
            token = self.auth.get_token()
            if token:
                headers["Authorization"] = "Firebase " + token
        return headers

    def _get_userid(self):
        userid = None
        if self.auth:
            userid = self.auth.get_user_id()
        userid = userid if userid else 'guest'
        return userid

    def _raise_detailed_error(self, request):
        if request.status_code != 200:
            raise ValueError(str(request.__dict__))

        data = request.json()
        if 'status' in data.keys():
            if data['status'] == 'ok':
                return

            raise ValueError(data['status'])
        else:
            raise ValueError(json.dumps(data))

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass



import unittest

from studio import runner


class RunnerTest(unittest.TestCase):

    def test_add_packages(self):

        list1 = ['keras==2.0.5', 'boto3==1.1.3']
        list2 = ['keras==1.0.9', 'h5py==2.7.0', 'abc']

        result = set(runner.add_packages(list1, list2))
        expected_result = set(['boto3==1.1.3', 'h5py==2.7.0',
                               'keras==1.0.9', 'abc'])

        self.assertEquals(result, expected_result)


if __name__ == '__main__':
    unittest.main()



import os
import uuid

import logging
import time
import tempfile
import re
from threading import Thread
import subprocess

import tarfile
try:
    from urllib.request import urlopen
except ImportError:
    from urllib import urlopen

import hashlib

from . import fs_tracker
from . import util
from .util import download_file, download_file_from_qualified, retry
from .util import compression_to_extension, compression_to_taropt, timeit
from .util import sixdecode

logging.basicConfig()


class TartifactStore(object):

    def __init__(self, measure_timestamp_diff=False, compression=None,
                 verbose=logging.DEBUG):
        if measure_timestamp_diff:
            try:
                self.timestamp_shift = self._measure_timestamp_diff()
            except BaseException:
                self.timestamp_shift = 0
        else:
            self.timestamp_shift = 0

        self.compression = compression

        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(verbose)

    def _measure_timestamp_diff(self):

        max_diff = 60

        tmpfile = os.path.join(
            tempfile.gettempdir(), str(
                uuid.uuid4()) + '.txt')
        with open(tmpfile, 'w') as f:
            f.write('timestamp_diff_test')
        key = 'tests/' + str(uuid.uuid4())
        self._upload_file(key, tmpfile)
        remote_timestamp = self._get_file_timestamp(key)

        if remote_timestamp is not None:

            now_remote_diff = time.time() - remote_timestamp
            self._delete_file(key)
            os.remove(tmpfile)

            assert -max_diff < now_remote_diff and \
                now_remote_diff < max_diff, \
                "Timestamp difference is more than 60 seconds. " + \
                "You'll need to adjust local clock for caching " + \
                "to work correctly"

            return -now_remote_diff if now_remote_diff < 0 else 0

    def get_artifact_hash(
            self,
            artifact,
            local_path=None,
            cache=True):

        if local_path is None:
            local_path = artifact['local']

        key = artifact.get('key')
        if os.path.exists(local_path):
            tar_filename = self._tartifact(local_path, key)
            retval = util.sha256_checksum(tar_filename)
            os.remove(tar_filename)
            return retval

    def put_artifact(
            self,
            artifact,
            local_path=None,
            cache=True,
            background=False):
        if local_path is None:
            local_path = artifact.get('local')

        key = artifact.get('key')

        if key and key.startswith('blobstore/') and \
           self._get_file_timestamp(key) is not None:
            self.logger.info(('Artifact with key {} exists in blobstore, ' +
                              'skipping the upload').format(key))

            return key

        if os.path.exists(local_path):
            tar_filename = self._tartifact(local_path, key)
            if key is None:
                key = 'blobstore/' + util.sha256_checksum(tar_filename) \
                      + '.tar' + compression_to_extension(self.compression)
                if self._get_file_timestamp(key) is not None:
                    self.logger.info(
                        ('Artifact with key {} exists in blobstore, ' +
                         'skipping the upload').format(key))

                    os.remove(tar_filename)
                    return key

            def finish_upload():
                self._upload_file(key, tar_filename)
                os.remove(tar_filename)

            if background:
                t = Thread(target=finish_upload)
                t.start()
                return (key, t)
            else:
                finish_upload()
                return key
        else:
            self.logger.debug(("Local path {} does not exist. " +
                               "Not uploading anything.").format(local_path))

    def get_artifact(
            self,
            artifact,
            local_path=None,
            only_newer=True,
            background=False):

        key = artifact.get('key')

        if key is None:
            assert not artifact['mutable']
            assert artifact.get('url') is not None or \
                artifact.get('qualified') is not None

            remote_path = artifact.get('url')
            if remote_path is None:
                remote_path = artifact.get('qualified')

            key = hashlib.sha256(remote_path.encode()).hexdigest()
            local_path = fs_tracker.get_blob_cache(key)
            if os.path.exists(local_path):
                self.logger.info((
                    'Immutable artifact exists at local_path {},' +
                    ' skipping the download').format(local_path))
                return local_path

            if artifact.get('url') is not None:
                download_file(remote_path, local_path, self.logger)
            else:
                if remote_path.startswith('dockerhub://') or \
                   remote_path.startswith('shub://'):
                    self.logger.info((
                        'Qualified {} points to a shub or dockerhub,' +
                        ' skipping the download'))
                    return remote_path

                download_file_from_qualified(
                    remote_path, local_path, self.logger)

        if local_path is None:
            if 'local' in artifact.keys() and \
                    os.path.exists(artifact['local']):
                local_path = artifact['local']
            else:
                if artifact['mutable']:
                    local_path = fs_tracker.get_artifact_cache(key)
                else:
                    local_path = fs_tracker.get_blob_cache(key)
                    if os.path.exists(local_path):
                        self.logger.info((
                            'Immutable artifact exists at local_path {},' +
                            ' skipping the download').format(local_path))
                        return local_path

        local_path = re.sub('\/\Z', '', local_path)
        local_basepath = os.path.dirname(local_path)

        self.logger.info("Downloading dir {} to local path {} from storage..."
                         .format(key, local_path))

        if only_newer and os.path.exists(local_path):
            self.logger.debug(
                'Comparing date of the artifact in storage with local')
            storage_time = self._get_file_timestamp(key)
            local_time = os.path.getmtime(local_path)
            if storage_time is None:
                self.logger.info(
                    "Unable to get storage timestamp, storage is either " +
                    "corrupted or has not finished uploading")
                return local_path

            if local_time > storage_time - self.timestamp_shift:
                self.logger.info(
                    "Local path is younger than stored, skipping the download")
                return local_path

        tar_filename = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        self.logger.debug("tar_filename = {} ".format(tar_filename))

        def finish_download():
            try:
                self._download_file(key, tar_filename)
            except BaseException as e:
                self.logger.debug(e)

            if os.path.exists(tar_filename):
                # first, figure out if the tar file has a base path of .
                # or not
                self.logger.info("Untarring {}".format(tar_filename))
                listtar, _ = subprocess.Popen(['tar', '-tf', tar_filename],
                                              stdout=subprocess.PIPE,
                                              stderr=subprocess.PIPE,
                                              close_fds=True
                                              ).communicate()
                listtar = listtar.strip().split(b'\n')
                listtar = [s.decode('utf-8') for s in listtar]

                self.logger.info('List of files in the tar: ' + str(listtar))
                if listtar[0].startswith('./'):
                    # Files are archived into tar from .; adjust path
                    # accordingly
                    basepath = local_path
                else:
                    basepath = local_basepath

                tarcmd = ('mkdir -p {} && ' +
                          'tar -xf {} -C {} --keep-newer-files') \
                    .format(basepath, tar_filename, basepath)

                self.logger.debug('Tar cmd = {}'.format(tarcmd))

                tarp = subprocess.Popen(
                    ['/bin/bash', '-c', tarcmd],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    close_fds=True)

                tarout, tarerr = tarp.communicate()
                if tarp.returncode != 0:
                    self.logger.info('tar had a non-zero return code!')
                    self.logger.info('tar cmd = ' + tarcmd)
                    self.logger.info('tar output: \n ' + str(tarout))

                if len(listtar) == 1:
                    actual_path = os.path.join(basepath, listtar[0])
                    self.logger.info(
                        'Renaming {} into {}'.format(
                            actual_path, local_path))
                    retry(lambda: os.rename(actual_path, local_path),
                          no_retries=5,
                          sleep_time=1,
                          exception_class=OSError,
                          logger=self.logger)

                os.remove(tar_filename)
            else:
                self.logger.warning(
                    'file {} download failed'.format(tar_filename))

        if background:
            t = Thread(target=finish_download)
            t.start()
            return (local_path, t)
        else:
            finish_download()
            return local_path

    @timeit
    def get_artifact_url(self, artifact, method='GET', get_timestamp=False):
        if 'key' in artifact.keys():
            url = self._get_file_url(artifact['key'], method=method)
        elif 'url' in artifact.keys():
            url = artifact['url']
        else:
            url = None

        if get_timestamp:
            timestamp = self._get_file_timestamp(artifact['key'])
            return (url, timestamp)
        else:
            return url

        return None

    def get_artifact_post(self, artifact):
        if 'key' in artifact.keys():
            return self._get_file_post(artifact['key'])
        return None

    def delete_artifact(self, artifact):
        if 'key' in artifact.keys():
            self._delete_file(artifact['key'])

    @timeit
    def stream_artifact(self, artifact):
        url = self.get_artifact_url(artifact)
        if url is None:
            return None

        fileobj = urlopen(url)
        if fileobj:
            try:
                retval = tarfile.open(fileobj=fileobj, mode='r|*')
                return retval
            except BaseException as e:
                fileobj.close()
                self.logger.info('Streaming artifact error:\n' + e.message)
        return None

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def _tartifact(self, local_path, key, cache=True):

        tar_filename = os.path.join(tempfile.gettempdir(),
                                    str(uuid.uuid4()))

        if os.path.isdir(local_path):
            local_basepath = local_path
            local_nameonly = '.'

        else:
            local_nameonly = os.path.basename(local_path)
            local_basepath = os.path.dirname(local_path)

        ignore_arg = ''
        ignore_filepath = os.path.join(local_basepath, ".studioml_ignore")
        if os.path.exists(ignore_filepath) and \
                not os.path.isdir(ignore_filepath):
            ignore_arg = "--exclude-from=%s" % ignore_filepath
            self.logger.debug('.studioml_ignore found: %s,'
                              ' files listed inside will'
                              ' not be tarred or uploaded'
                              % ignore_filepath)

        if cache and key:
            cache_dir = fs_tracker.get_artifact_cache(key)
            if cache_dir != local_path:
                debug_str = "Copying local path {} to cache {}" \
                    .format(local_path, cache_dir)
                if ignore_arg != '':
                    debug_str += ", excluding files in {}" \
                        .format(ignore_filepath)
                self.logger.debug(debug_str)

                util.rsync_cp(local_path, cache_dir, ignore_arg,
                              self.logger)

        debug_str = ("Tarring artifact. " +
                     "tar_filename = {}, " +
                     "local_path = {}, " +
                     "key = {}").format(
            tar_filename,
            local_path,
            key)

        if ignore_arg != '':
            debug_str += ", exclude = {}".format(ignore_filepath)
        self.logger.debug(debug_str)

        tarcmd = 'tar {} {} -cf {} -C {} {}'.format(
            ignore_arg,
            compression_to_taropt(self.compression),
            tar_filename,
            local_basepath,
            local_nameonly)
        self.logger.debug("Tar cmd = {}".format(tarcmd))

        tic = time.time()
        tarp = subprocess.Popen(['/bin/bash', '-c', tarcmd],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                close_fds=True)

        tarout, _ = tarp.communicate()
        toc = time.time()

        if tarp.returncode != 0:
            self.logger.info('tar had a non-zero return code!')
            self.logger.info('tar output: \n ' + sixdecode(tarout))

        self.logger.info('tar finished in {}s'.format(toc - tic))
        return tar_filename


def get_immutable_artifact_key(arthash, compression=None):
    retval = "blobstore/" + arthash + ".tar" + \
             compression_to_extension(compression)
    return retval



import cma
import random
import copy
import pickle
# import cPickle as pickle
import os
import pprint
import time
import six

import numpy as np

from opt_util import EPSILON


class Optimizer(object):

    def __init__(self, hyperparameters, config, logger):
        self.hyperparameters = hyperparameters
        self.config = config
        self.logger = logger

        self.opts = cma.CMAOptions()
        for param, value in six.iteritems(self.config['cmaes_config']):
            if param in self.opts and value is not None:
                self.opts[param] = value
        self.dim = 0
        for h in self.hyperparameters:
            assert self.dim == h.index
            self.dim += h.array_length if h.array_length is not None else 1

        self.init = np.empty(self.dim)
        # self.sigma = np.random.random(self.dim) # not allowed
        self.sigma = self.config['cmaes_config']['sigma0']
        self.opts['CMA_stds'] = np.ones(self.dim)
        self.gen = 0
        self.start_time = time.time()
        self.best_fitnesses = []
        self.mean_fitnesses = []
        self.best = None

        for h in self.hyperparameters:

            if h.array_length is None:
                if h.rand_init:
                    self.init[h.index] = random.random(
                    ) * (h.max_range - h.min_range) + h.min_range
                else:
                    self.init[h.index] = (h.max_range + h.min_range) / 2.0
            else:
                if h.rand_init:
                    self.init[h.index: h.index + h.array_length] = \
                        np.random.random(h.array_length) * \
                        (h.max_range - h.min_range) + h.min_range
                else:
                    self.init[h.index: h.index + h.array_length] = \
                        np.ones(h.array_length) * (h.max_range + h.min_range) \
                        / 2.0

            if h.array_length is None:
                if h.max_range - h.min_range > EPSILON:
                    self.opts['CMA_stds'][h.index] *= h.max_range - h.min_range
            else:
                if h.max_range - h.min_range > EPSILON:
                    self.opts['CMA_stds'][
                        h.index: h.index + h.array_length] *= \
                        h.max_range - h.min_range

        # If min range and max range are exactly the same,
        # use a sigma calculated
        # from mean of init
        # if max([h.max_range for h in hyperparameters]) - \
        #     min([h.min_range for h in hyperparameters]) < EPSILON:
        #     self.logger.warn("min range == max range, overwriting sigma0")
        #     self.sigma = np.mean(self.init) * MISC_CONFIG['sigma0']
        self.es = cma.CMAEvolutionStrategy(self.init, self.sigma, self.opts)
        self.__load_checkpoint()

        self.logger.info("Init: %s" % self.init)
        self.logger.info("CMA stds: %s" % self.opts['CMA_stds'])
        self.logger.info(pprint.pformat(self.get_config()))

    best_fitness = property(lambda self: self.best_fitnesses[-1]
                            if len(self.best_fitnesses) > 0 else 0.0)

    mean_fitness = property(lambda self: self.mean_fitnesses[-1]
                            if len(self.mean_fitnesses) > 0 else 0.0)

    def get_config(self):
        return self.config['cmaes_config']

    def __unpack_solution(self, solution):
        # print solution
        new_hyperparameters = []
        for h in self.hyperparameters:
            h = copy.copy(h)
            if h.array_length is None:
                h.values = solution[h.index]
            else:
                h.values = solution[h.index: h.index + h.array_length]
            if not h.unbounded:
                h.values = np.clip(h.values, h.min_range, h.max_range)
            # if h.max_range - h.min_range < EPSILON:
            #     h.values = unscale_var(h.values, h.min_range, h.max_range)
            if h.is_log:
                h.values = np.exp(h.values)
            if h.array_length is None:
                h.values = float(h.values)
            new_hyperparameters.append(h)
        return new_hyperparameters

    def __pack_solution(self, hyperparameters):
        solution = np.empty(self.dim)
        for h in hyperparameters:
            values = copy.copy(h.values)
            if h.is_log:
                values = np.log(values)
            # if h.max_range - h.min_range < EPSILON:
            #     values = scale_var(values, h.min_range, h.max_range)
            if not h.unbounded:
                values = np.clip(values, h.min_range, h.max_range)
            if h.array_length is None:
                solution[h.index] = values
            else:
                solution[h.index: h.index + h.array_length] = values
        # print solution
        return solution

    def stop(self):
        term_criterion = self.config['termination_criterion']

        if self.gen >= term_criterion['generation']:
            self.logger.info("Reached target generation %s, terminating" %
                             term_criterion['generation'])
            return True
        elif self.best_fitness >= term_criterion['fitness']:
            self.logger.info("Reached target fitness %s, terminating" %
                             term_criterion['fitness'])
            return True
        return False
        # return self.es.stop()

    def ask(self):
        solutions = self.es.ask()
        return [self.__unpack_solution(s) for s in solutions]

    def tell(self, hyperparameter_pop, fitnesses):
        adjusted_fitnesses = -1 * np.array(fitnesses)
        self.best_fitnesses.append(float(np.max(fitnesses)))
        self.mean_fitnesses.append(float(np.mean(fitnesses)))
        self.best = (hyperparameter_pop[np.argmax(fitnesses)],
                     self.__pack_solution(
            hyperparameter_pop[np.argmax(fitnesses)]))

        solutions = [self.__pack_solution(hyperparameters) for hyperparameters
                     in hyperparameter_pop]
        self.es.tell(solutions, adjusted_fitnesses)
        self.gen += 1
        self.__save_checkpoint()

    def disp(self):
        print("**************************************************************")
        print("CMAES wall time: %s gen: %s pop size: %s best fitness: "
              "%s mean fitness: %s" % (int(time.time() - self.start_time),
                                       self.gen, self.es.popsize,
                                       self.best_fitness, self.mean_fitness))
        print("**************************************************************")

    def __load_checkpoint(self):
        if self.config['load_checkpoint_file'] is None:
            return
        try:
            checkpoint_file = os.path.abspath(os.path.expanduser(
                self.config['load_checkpoint_file']))
            with open(checkpoint_file) as f:
                old_cmaes_instance = pickle.load(f)
        except BaseException:
            self.logger.warn("Checkpoint file cannot be loaded")
            raise

        for h, h_old in zip(self.hyperparameters,
                            old_cmaes_instance.hyperparameters):
            assert h.is_compatible(h_old)

        if self.config['cmaes_config']['load_best_only']:
            self.init = old_cmaes_instance.best[1]
            self.es = cma.CMAEvolutionStrategy(self.init, self.sigma,
                                               self.opts)
            self.logger.info("Loaded best solution as init")
        else:
            logger = self.logger
            config = self.config
            self.__dict__ = old_cmaes_instance.__dict__
            self.logger = logger
            self.config = config

        self.logger.info("Loaded checkpoint from file: %s" % checkpoint_file)

    def __save_checkpoint(self):
        if (int(self.config['checkpoint_interval']) >= 1 and
            self.gen % int(self.config['checkpoint_interval']) == 0) or \
                self.stop():

            try:
                result_dir = os.path.abspath(
                    os.path.expanduser(self.config['result_dir']))
                if not os.path.exists(result_dir):
                    os.makedirs(result_dir)
            except BaseException:
                self.looger.warn("Cannot retrieve checkpoint directory,"
                                 " not saving checkpoint")
                return

            checkpoint_file = os.path.join(
                result_dir, "G%s_F%s_checkpoint.pkl" %
                (self.gen, self.best_fitness))
            with open(checkpoint_file, 'wb') as f:
                copy_of_self = copy.copy(self)
                copy_of_self.logger = None
                copy_of_self.config = None
                # copy_of_self.es = None
                pickle.dump(copy_of_self, f, protocol=-1)

            best_file = os.path.join(
                result_dir, "G%s_F%s_best.pkl" %
                (self.gen, self.best_fitness))
            with open(best_file, 'wb') as f:
                pickle.dump(self.best, f, protocol=-1)

            with open(os.path.join(result_dir, "fitness.txt"), 'wb') as f:
                for best, mean in zip(
                        self.best_fitnesses, self.mean_fitnesses):
                    f.write("%s %s\n" % (best, mean))

            self.logger.info("Saved checkpoint to file: %s" % checkpoint_file)



import logging
import time
import os
import six
import re
from threading import Thread

from . import util, git_util, pyrebase
from .firebase_artifact_store import FirebaseArtifactStore
from .auth import get_auth
from .experiment import experiment_from_dict
from .tartifact_store import get_immutable_artifact_key
from .util import timeit

logging.basicConfig()


class KeyValueProvider(object):
    """Data provider for Firebase."""

    def __init__(
            self,
            db_config,
            blocking_auth=True,
            verbose=10,
            store=None,
            compression=None):
        guest = db_config.get('guest')

        self.app = pyrebase.initialize_app(db_config)
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(verbose)

        self.compression = compression
        if self.compression is None:
            self.compression = db_config.get('compression')

        self.auth = None
        if not guest and 'serviceAccount' not in db_config.keys():
            self.auth = get_auth(self.app,
                                 db_config.get("use_email_auth"),
                                 db_config.get("email"),
                                 db_config.get("password"),
                                 blocking_auth)

        self.store = store if store else FirebaseArtifactStore(
            db_config,
            verbose=verbose,
            blocking_auth=blocking_auth,
            compression=self.compression
        )

        if self.auth and not self.auth.expired:
            self.register_user(None, self.auth.get_user_email())

        self.max_keys = db_config.get('max_keys', 100)

    def _get_userid(self):
        userid = None
        if self.auth:
            userid = self.auth.get_user_id()
        userid = userid if userid else 'guest'
        return userid

    def _get_user_keybase(self, userid=None):
        if userid is None:
            userid = self._get_userid()

        return "users/" + userid + "/"

    def _get_experiments_keybase(self, userid=None):
        return "experiments/"

    def _get_projects_keybase(self):
        return "projects/"

    def add_experiment(self, experiment, userid=None, compression=None):
        self._delete(self._get_experiments_keybase() + experiment.key)
        experiment.time_added = time.time()
        experiment.status = 'waiting'

        compression = compression if compression else self.compression

        if 'local' in experiment.artifacts['workspace'].keys() and \
                os.path.exists(experiment.artifacts['workspace']['local']):
            experiment.git = git_util.get_git_info(
                experiment.artifacts['workspace']['local'])

        for tag, art in six.iteritems(experiment.artifacts):
            if art['mutable']:
                art['key'] = self._get_experiments_keybase() + \
                    experiment.key + '/' + tag + '.tar' + \
                    util.compression_to_extension(compression)
            else:
                if 'local' in art.keys():
                    # upload immutable artifacts
                    art['key'] = self.store.put_artifact(art)
                elif 'hash' in art.keys():
                    art['key'] = get_immutable_artifact_key(
                        art['hash'],
                        compression=compression
                    )

            key = art.get('key')
            if key is not None:
                art['qualified'] = self.store.get_qualified_location(key)
                art['bucket'] = self.store.get_bucket()
            elif art.get('qualified'):
                qualified = art.get('qualified')
                bucket = re.search('(?<=://)[^/]+(?=/)', qualified).group(0)
                if bucket.endswith('.com'):
                    bucket = re.search(
                        '(?<=' + re.escape(bucket) + '/)[^/]+(?=/)',
                        qualified
                    ).group(0)

                key = re.search('(?<=' + bucket + '/).+\Z', qualified).group(0)
                art['bucket'] = bucket
                art['key'] = key

        userid = userid if userid else self._get_userid()
        experiment.owner = userid

        experiment_dict = experiment.__dict__.copy()

        self._set(self._get_experiments_keybase() + experiment.key,
                  experiment_dict)

        self._set(self._get_user_keybase(userid) + "experiments/" +
                  experiment.key,
                  experiment.time_added)

        if experiment.project and userid:
            self._set(self._get_projects_keybase() +
                      experiment.project + "/" +
                      experiment.key + "/owner",
                      userid)

        self.checkpoint_experiment(experiment, blocking=True)
        self.logger.info("Added experiment " + experiment.key)

    def start_experiment(self, experiment):
        time_started = time.time()
        experiment.time_started = time_started
        experiment.status = 'running'

        experiment_dict = self._get(self._get_experiments_keybase() +
                                    experiment.key)
        experiment_dict['time_started'] = time_started
        experiment_dict['status'] = 'running'

        self._set(self._get_experiments_keybase() +
                  experiment.key,
                  experiment_dict)

        self.checkpoint_experiment(experiment)

    def stop_experiment(self, key):
        # can be called remotely (the assumption is
        # that remote worker checks experiments status periodically,
        # and if it is 'stopped', kills the experiment)
        if not isinstance(key, six.string_types):
            key = key.key

        experiment_data = self._get(self._get_experiments_keybase() +
                                    key)

        experiment_data['status'] = 'stopped'

        self._set(self._get_experiments_keybase() +
                  key, experiment_data)

    def finish_experiment(self, experiment):
        time_finished = time.time()
        if not isinstance(experiment, six.string_types):
            key = experiment.key
            experiment.status = 'finished'
            experiment.time_finished = time_finished
        else:
            key = experiment

        experiment_dict = self._get(
            self._get_experiments_keybase() + key)

        experiment_dict['status'] = 'finished'
        experiment_dict['time_finished'] = time_finished

        self._set(self._get_experiments_keybase() +
                  key, experiment_dict)

    def delete_experiment(self, experiment):
        if isinstance(experiment, six.string_types):
            experiment_key = experiment
            try:
                experiment = self.get_experiment(experiment)
                experiment_key = experiment.key
            except BaseException:
                experiment = None
        else:
            experiment_key = experiment.key

        experiment_owner = self._get(self._get_experiments_keybase() +
                                     experiment_key).get('owner')

        self._delete(self._get_user_keybase(experiment_owner) +
                     'experiments/' + experiment_key)
        if experiment is not None:
            for tag, art in six.iteritems(experiment.artifacts):
                if art.get('key') is not None:
                    self.logger.debug(
                        ('Deleting artifact {} from the store, ' +
                         'artifact key {}').format(tag, art['key']))
                    self.store.delete_artifact(art)

            if experiment.project is not None:
                self._delete(
                    self._get_projects_keybase() +
                    experiment.project +
                    "/" +
                    experiment_key)

        self._delete(self._get_experiments_keybase() + experiment_key)

    def checkpoint_experiment(self, experiment, blocking=False):
        if not isinstance(experiment, six.string_types):
            key = experiment.key
        else:
            key = experiment

        experiment_dict = self._get(self._get_experiments_keybase() + key)

        checkpoint_threads = [
            Thread(
                target=self.store.put_artifact,
                args=(art,))
            for _, art in six.iteritems(experiment.artifacts)
            if art['mutable'] and art.get('local')]

        for t in checkpoint_threads:
            t.start()

        checkpoint_time = time.time()
        experiment.time_last_checkpoint = checkpoint_time
        experiment_dict['time_last_checkpoint'] = checkpoint_time

        self._set(self._get_experiments_keybase() +
                  key, experiment_dict)

        if blocking:
            for t in checkpoint_threads:
                t.join()
        else:
            return checkpoint_threads

    @timeit
    def _get_experiment_info(self, experiment):
        info = {}
        type_found = False

        if not type_found:
            info['type'] = 'unknown'

        info['logtail'] = self._get_experiment_logtail(experiment)

        if experiment.metric is not None:
            metric_str = experiment.metric.split(':')
            metric_name = metric_str[0]
            metric_type = metric_str[1] if len(metric_str) > 1 else None

            tbtar = self.store.stream_artifact(experiment.artifacts['tb'])

            if metric_type == 'min':
                def metric_accum(x, y): return min(x, y) if x else y
            elif metric_type == 'max':
                def metric_accum(x, y): return max(x, y) if x else y
            else:
                def metric_accum(x, y): return y

            metric_value = None
            for f in tbtar:
                if f.isreg():
                    for e in util.event_reader(tbtar.extractfile(f)):
                        for v in e.summary.value:
                            if v.tag == metric_name:
                                metric_value = metric_accum(
                                    metric_value, v.simple_value)

            info['metric_value'] = metric_value

        return info

    @timeit
    def _get_experiment_logtail(self, experiment):
        try:
            tarf = self.store.stream_artifact(experiment.artifacts['output'])
            if not tarf:
                return None
            tarf_member = tarf.members[0]
            while tarf_member is not None and tarf_member.name == '':
                tarf_member = tarf.next()

            logdata = tarf.extractfile(tarf_member).read()
            logdata = util.remove_backspaces(logdata).split('\n')
            return logdata
        except BaseException as e:
            self.logger.info('Getting experiment logtail raised an exception:')
            self.logger.info(e)
            return None

    @timeit
    def get_experiment(self, key, getinfo=True):
        data = self._get(self._get_experiments_keybase() + key)
        if data is None:
            return None

        data['key'] = key

        experiment_stub = experiment_from_dict(data)

        expinfo = {}
        if getinfo:
            try:
                expinfo = self._get_experiment_info(experiment_stub)

            except Exception as e:
                self.logger.info(
                    "Exception {} while info download for {}".format(
                        e, key))

        return experiment_from_dict(data, expinfo)

    def get_user_experiments(self, userid=None, blocking=True):
        if userid and '@' in userid:
            users = self.get_users()
            user_ids = [u for u in users if users[u].get('email') == userid]
            if len(user_ids) < 1:
                return None
            else:
                userid = user_ids[0]

        experiment_keys = self._get(
            self._get_user_keybase(userid) + "experiments/", shallow=True)
        if not experiment_keys:
            experiment_keys = []

        # keys = sorted(experiment_keys.keys(),
        #              key=lambda k: str(experiment_keys[k]),
        #              reverse=True)

        return experiment_keys

    def get_project_experiments(self, project):
        experiment_keys = self._get(self._get_projects_keybase() +
                                    project)
        if not experiment_keys:
            experiment_keys = {}

        return experiment_keys.keys()

    def get_artifacts(self, key):
        if isinstance(key, six.string_types):
            experiment = self.get_experiment(key, getinfo=False)
        else:
            experiment = key

        retval = {}
        if experiment.artifacts is not None:
            for tag, art in six.iteritems(experiment.artifacts):
                url = self.store.get_artifact_url(art)
                if url is not None:
                    retval[tag] = url

        return retval

    def get_artifact(self, artifact, local_path=None, only_newer=True):
        return self.store.get_artifact(artifact,
                                       local_path=local_path,
                                       only_newer=only_newer)

    def get_projects(self):
        return self._get(self._get_projects_keybase(), shallow=True)

    def get_users(self):
        user_ids = self._get('users/', shallow=True)
        retval = {}
        if user_ids:
            for user_id in user_ids:
                retval[user_id] = {
                    'email': self._get('users/' + user_id + '/email')
                }
        return retval

    def refresh_auth_token(self, email, refresh_token):
        if self.auth:
            self.auth.refresh_token(email, refresh_token)

    def is_auth_expired(self):
        if self.auth:
            return self.auth.expired
        else:
            return False

    def can_write_experiment(self, key=None, user=None):
        assert key is not None
        user = user if user else self._get_userid()

        experiment = self._get(
            self._get_experiments_keybase() + key)

        if experiment:
            owner = experiment.get('owner')
            if owner is None or owner == 'guest':
                return True
            else:
                return (owner == user)
        else:
            return True

    def register_user(self, userid, email):
        keypath = self._get_user_keybase(userid) + 'email'
        existing_email = self._get(keypath)
        if existing_email != email:
            self._set(keypath, email)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if self.app:
            self.app.requests.close()
        if self.store:
            self.store.__exit__()



import unittest
import uuid
import pip
import os

from studio import model
from studio.experiment import create_experiment


def get_test_experiment():
    filename = 'test.py'
    args = ['a', 'b', 'c']
    experiment_name = 'test_experiment_' + str(uuid.uuid4())
    experiment = create_experiment(filename, args, experiment_name)
    return experiment, experiment_name, filename, args


class ModelTest(unittest.TestCase):
    def test_create_experiment(self):
        _, experiment_name, filename, args = get_test_experiment()
        experiment_project = 'create_experiment_project'
        experiment = create_experiment(
            filename, args, experiment_name, experiment_project)
        packages = [
            p._key +
            '==' +
            p._version for p in pip.pip.get_installed_distributions(
                local_only=True) if p._key != 'studioml']

        self.assertTrue(experiment.key == experiment_name)
        self.assertTrue(experiment.filename == filename)
        self.assertTrue(experiment.args == args)
        self.assertTrue(experiment.project == experiment_project)
        self.assertTrue(sorted(experiment.pythonenv) == sorted(packages))

    def test_get_config_env(self):
        value1 = str(uuid.uuid4())
        os.environ['TEST_VAR1'] = value1
        value2 = str(uuid.uuid4())
        os.environ['TEST_VAR2'] = value2

        config = model.get_config(
            os.path.join(os.path.dirname(os.path.realpath(__file__)),
                         'test_config_env.yaml'))
        self.assertEquals(config['test_key'], value1)
        self.assertEquals(config['test_section']['test_key'], value2)


if __name__ == "__main__":
    unittest.main()



import unittest
import os
import uuid

from studio.gcloud_worker import GCloudWorkerManager
from studio.ec2cloud_worker import EC2WorkerManager
from local_worker_test import stubtest_worker

from timeout_decorator import timeout
from studio.util import has_aws_credentials


@unittest.skipIf(
    'GOOGLE_APPLICATION_CREDENTIALS' not in os.environ.keys(),
    'GOOGLE_APPLICATION_CREDENTIALS environment ' +
    'variable not set, won'' be able to use google cloud')
class GCloudWorkerTest(unittest.TestCase):
    _multiprocess_shared_ = True

    def get_worker_manager(self):
        project = 'studio-ed756'
        return GCloudWorkerManager(project)

    @timeout(600, use_signals=False)
    def test_worker(self):
        experiment_name = 'test_gcloud_worker_' + str(uuid.uuid4())
        with stubtest_worker(
            self,
            experiment_name=experiment_name,
            runner_args=['--cloud=gcloud', '--force-git',
                         '--cloud-timeout=120'],
            config_name='test_config_http_client.yaml',
            test_script='tf_hello_world.py',
            script_args=['arg0'],
            expected_output='[ 2.  6.]',
        ):
            pass

    @timeout(600, use_signals=False)
    def test_worker_spot(self):
        experiment_name = 'test_gcloud_spot_worker_' + str(uuid.uuid4())
        with stubtest_worker(
            self,
            experiment_name=experiment_name,
            runner_args=['--cloud=gcspot', '--force-git',
                         '--cloud-timeout=120'],
            config_name='test_config_http_client.yaml',
            test_script='tf_hello_world.py',
            script_args=['arg0'],
            expected_output='[ 2.  6.]',
        ):
            pass

    @timeout(600, use_signals=False)
    def test_worker_spot_container(self):
        experiment_name = 'test_gcloud_spot_simg_' + str(uuid.uuid4())
        with stubtest_worker(
            self,
            experiment_name=experiment_name,
            runner_args=['--cloud=gcspot',
                         '--force-git',
                         '--cloud-timeout=120',
                         '--container=shub://vsoch/hello-world'],

            config_name='test_config_http_client.yaml',
            test_script='',
            script_args=[],
            expected_output='RaawwWWWWWRRRR!!',
            test_workspace=False
        ):
            pass


@unittest.skipIf(
    not has_aws_credentials(),
    'boto3 not present, won\'t be able to use AWS API')
class EC2WorkerTest(unittest.TestCase):
    _multiprocess_shared_ = True

    def get_worker_manager(self):
        return EC2WorkerManager()

    @timeout(600, use_signals=False)
    def test_worker(self):
        experiment_name = 'test_ec2_worker_' + str(uuid.uuid4())
        with stubtest_worker(
            self,
            experiment_name=experiment_name,
            runner_args=['--cloud=ec2', '--force-git',
                         '--cloud-timeout=120'],
            config_name='test_config_http_client.yaml',
            test_script='tf_hello_world.py',
            script_args=['arg0'],
            expected_output='[ 2.  6.]',
        ):
            pass

    @timeout(600, use_signals=False)
    def test_worker_spot(self):
        experiment_name = 'test_ec2_worker_' + str(uuid.uuid4())
        stubtest_worker(
            self,
            experiment_name=experiment_name,
            runner_args=[
                '--cloud=ec2spot',
                '--force-git',
                '--bid=50%',
                '--cloud-timeout=120',
            ],
            config_name='test_config_http_client.yaml',
            test_script='tf_hello_world.py',
            script_args=['arg0'],
            expected_output='[ 2.  6.]',
        )

    def test_get_ondemand_prices(self):
        wm = self.get_worker_manager()
        prices = wm._get_ondemand_prices(['c4.large', 'p2.xlarge'])

        expected_prices = {'c4.large': 0.1, 'p2.xlarge': 0.9}
        self.assertEquals(prices, expected_prices)

    @timeout(600, use_signals=False)
    def test_worker_spot_container(self):
        experiment_name = 'test_gcloud_spot_simg_' + str(uuid.uuid4())
        with stubtest_worker(
            self,
            experiment_name=experiment_name,
            runner_args=['--cloud=ec2spot',
                         '--force-git',
                         '--cloud-timeout=120',
                         '--container=shub://vsoch/hello-world'],

            config_name='test_config_http_client.yaml',
            test_script='',
            script_args=[],
            expected_output='RaawwWWWWWRRRR!!',
            test_workspace=False
        ):
            pass


if __name__ == '__main__':
    unittest.main()



"""Tools to save/restore model from checkpoints."""

import os
try:
    import torch
except ImportError:
    torch = None


def load_checkpoint(model, optimizer, model_dir, map_to_cpu=False):
    path = os.path.join(model_dir, 'checkpoint')
    if os.path.exists(path):
        print("Loading model from %s" % path)
        if map_to_cpu:
            checkpoint = torch.load(
                path, map_location=lambda storage, location: storage)
        else:
            checkpoint = torch.load(path)
        old_state_dict = model.state_dict()
        for key in old_state_dict.keys():
            if key not in checkpoint['model']:
                checkpoint['model'][key] = old_state_dict[key]
        model.load_state_dict(checkpoint['model'])
        optimizer.load_state_dict(checkpoint['optimizer'])
        return checkpoint.get('step', 0)
    return 0


def save_checkpoint(model, optimizer, step, model_dir, ignore=[]):
    if not os.path.exists(model_dir):
        os.makedirs(model_dir)
    path = os.path.join(model_dir, 'checkpoint')
    state_dict = model.state_dict()
    if ignore:
        for key in state_dict.keys():
            for item in ignore:
                if key.startswith(item):
                    state_dict.pop(key)
    torch.save({
        'model': state_dict,
        'optimizer': optimizer.state_dict(),
        'step': step
    }, path)


class Saver(object):
    """Class to manage save and restore for the model and optimizer."""

    def __init__(self, model, optimizer):
        self._model = model
        self._optimizer = optimizer

    def restore(self, model_dir, map_to_cpu=False):
        """Restores model and optimizer from given directory.

        Returns:
           Last training step for the model restored.
        """
        last_step = load_checkpoint(
            self._model, self._optimizer, model_dir, map_to_cpu)
        return last_step

    def save(self, model_dir, step):
        """Saves model and optimizer to given directory.

        Args:
           model_dir: Model directory to save.
           step: Current training step.
        """
        save_checkpoint(self._model, self._optimizer, step, model_dir)



import unittest
import os
import tempfile
import uuid
import subprocess

from studio.pubsub_queue import PubsubQueue
from local_worker_test import stubtest_worker

from timeout_decorator import timeout
import logging

logging.basicConfig()


@unittest.skip('testing requires docker')
class RemoteWorkerTest(unittest.TestCase):
    _multiprocess_shared_ = True

    @timeout(590)
    @unittest.skipIf(
        'GOOGLE_APPLICATION_CREDENTIALS' not in
        os.environ.keys(),
        'GOOGLE_APPLICATION_CREDENTIALS environment ' +
        'variable not set, won'' be able to use google ' +
        'PubSub')
    def test_remote_worker(self):
        experiment_name = 'test_remote_worker_' + str(uuid.uuid4())
        queue_name = experiment_name
        logger = logging.getLogger('test_remote_worker')
        logger.setLevel(10)

        pw = subprocess.Popen(
            ['studio-start-remote-worker',
             '--queue=' + queue_name,
             '--single-run',
             '--no-cache',
             '--timeout=30',
             '--image=peterzhokhoff/studioml'],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT
        )

        stubtest_worker(
            self,
            experiment_name=experiment_name,
            runner_args=['--queue=' + queue_name, '--force-git'],
            config_name='test_config_http_client.yaml',
            test_script='tf_hello_world.py',
            script_args=['arg0'],
            expected_output='[ 2.  6.]',
            queue=PubsubQueue(queue_name))

        workerout, _ = pw.communicate()
        if workerout:
            logger.debug("studio-start-remote-worker output: \n" +
                         str(workerout))

    @timeout(590)
    @unittest.skipIf(
        'GOOGLE_APPLICATION_CREDENTIALS' not in
        os.environ.keys(),
        'GOOGLE_APPLICATION_CREDENTIALS environment ' +
        'variable not set, won'' be able to use google ' +
        'PubSub')
    def test_remote_worker_c(self):
        tmpfile = os.path.join(tempfile.gettempdir(),
                               str(uuid.uuid4()))

        logger = logging.getLogger('test_remote_worker_c')
        logger.setLevel(10)
        experiment_name = "test_remote_worker_c_" + str(uuid.uuid4())

        random_str1 = str(uuid.uuid4())
        with open(tmpfile, 'w') as f:
            f.write(random_str1)

        random_str2 = str(uuid.uuid4())

        queue_name = experiment_name
        pw = subprocess.Popen(
            ['studio-start-remote-worker',
             '--queue=' + queue_name,
             '--single-run',
             '--no-cache',
             '--image=peterzhokhoff/studioml'],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)

        db = stubtest_worker(
            self,
            experiment_name=experiment_name,
            runner_args=[
                '--capture=' + tmpfile + ':f',
                '--queue=' + queue_name,
                '--force-git'],
            config_name='test_config_http_client.yaml',
            test_script='art_hello_world.py',
            script_args=[random_str2],
            expected_output=random_str1,
            queue=PubsubQueue(queue_name),
            delete_when_done=False)

        workerout, _ = pw.communicate()
        if workerout:
            logger.debug("studio-start-remote-worker output: \n" +
                         str(workerout))
        os.remove(tmpfile)

        tmppath = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        if os.path.exists(tmppath):
            os.remove(tmppath)

        db.get_artifact(
            db.get_experiment(experiment_name).artifacts['f'],
            tmppath,
            only_newer=False
        )

        with open(tmppath, 'r') as f:
            self.assertEquals(f.read(), random_str2)
        os.remove(tmppath)
        db.delete_experiment(experiment_name)

    @timeout(590)
    @unittest.skipIf(
        'GOOGLE_APPLICATION_CREDENTIALS' not in
        os.environ.keys(),
        'GOOGLE_APPLICATION_CREDENTIALS environment ' +
        'variable not set, won'' be able to use google ' +
        'PubSub')
    def test_remote_worker_co(self):
        logger = logging.getLogger('test_remote_worker_co')
        logger.setLevel(10)

        tmpfile = os.path.join(tempfile.gettempdir(),
                               str(uuid.uuid4()))

        random_str = str(uuid.uuid4())
        with open(tmpfile, 'w') as f:
            f.write(random_str)

        experiment_name = 'test_remote_worker_co_' + str(uuid.uuid4())
        queue_name = experiment_name
        pw = subprocess.Popen(
            ['studio-start-remote-worker',
             '--queue=' + queue_name,
             '--single-run',
             '--no-cache',
             '--image=peterzhokhoff/studioml'],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)

        stubtest_worker(
            self,
            experiment_name=experiment_name,
            runner_args=[
                '--capture-once=' + tmpfile + ':f',
                '--queue=' + queue_name,
                '--force-git'],
            config_name='test_config_http_client.yaml',
            test_script='art_hello_world.py',
            script_args=[],
            expected_output=random_str,
            queue=PubsubQueue(queue_name))

        workerout, _ = pw.communicate()
        logger.debug('studio-start-remote-worker output: \n' + str(workerout))

        os.remove(tmpfile)

    @timeout(590)
    @unittest.skipIf(
        'GOOGLE_APPLICATION_CREDENTIALS' not in
        os.environ.keys(),
        'GOOGLE_APPLICATION_CREDENTIALS environment ' +
        'variable not set, won'' be able to use google ' +
        'PubSub')
    def test_baked_image(self):

        # create a docker image with baked in credentials
        # and run a remote worker tests with it
        logger = logging.getLogger('test_baked_image')
        logger.setLevel(logging.DEBUG)

        # check if docker is installed
        dockertestp = subprocess.Popen(['docker'],
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.STDOUT)

        dockertestout, _ = dockertestp.communicate()
        if dockertestout:
            logger.info("docker test output: \n" + str(dockertestout))

        if dockertestp.returncode != 0:
            logger.error("docker is not installed (correctly)")
            return

        image = 'test_image' + str(uuid.uuid4())

        addcredsp = subprocess.Popen(
            [
                'studio-add-credentials',
                '--tag=' + image,
                '--base-image=peterzhokhoff/studioml'],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)

        addcredsout, _ = addcredsp.communicate()
        if addcredsout:
            logger.info('studio-add-credentials output: \n' + str(addcredsout))
        if addcredsp.returncode != 0:
            logger.error("studio-add-credentials failed.")
            self.assertTrue(False)

        experiment_name = 'test_remote_worker_baked' + str(uuid.uuid4())
        queue_name = experiment_name
        logger = logging.getLogger('test_baked_image')
        logger.setLevel(10)

        pw = subprocess.Popen(
            ['studio-start-remote-worker',
             '--queue=' + queue_name,
             '--no-cache',
             '--single-run',
             '--timeout=30',
             '--image=' + image],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)

        stubtest_worker(
            self,
            experiment_name=experiment_name,
            runner_args=['--queue=' + queue_name, '--force-git'],
            config_name='test_config_http_client.yaml',
            test_script='tf_hello_world.py',
            script_args=['arg0'],
            expected_output='[ 2.  6.]',
            queue=PubsubQueue(queue_name))

        workerout, _ = pw.communicate()
        if workerout:
            logger.debug(
                "studio-start-remote-worker output: \n" +
                str(workerout))

        rmip = subprocess.Popen(['docker', 'rmi', image],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)

        rmiout, _ = rmip.communicate()

        if rmiout:
            logger.info('docker rmi output: \n' + str(rmiout))


if __name__ == "__main__":
    unittest.main()



import uuid
import unittest
import os
import logging

from .completion_service import CompletionService

from studio.util import has_aws_credentials
from studio.local_queue import get_local_queue_lock


logging.basicConfig()


class CompletionServiceTest(unittest.TestCase):

    _multiprocess_shared_ = True

    def test_two_experiments_with_cs_args(self, n_experiments=2, **kwargs):
        if not(any(kwargs)):
            return
        mypath = os.path.dirname(os.path.realpath(__file__))
        experimentId = str(uuid.uuid4())
        results = {}
        expected_results = {}
        with CompletionService(experimentId, **kwargs) as cs:
            for i in range(0, n_experiments):
                key = cs.submitTask(
                    os.path.join(
                        mypath,
                        'completion_service_func.py'),
                    [i])
                expected_results[key] = [i]

            for i in range(0, n_experiments):
                result = cs.getResults(blocking=True)
                results[result[0]] = result[1]

        self.assertEquals(results, expected_results)

    @unittest.skipIf(not has_aws_credentials(),
                     'AWS credentials needed for this test')
    def test_two_experiments_ec2(self):
        mypath = os.path.dirname(os.path.realpath(__file__))
        config_path = os.path.join(
            mypath,
            '..',
            'tests',
            'test_config_http_client.yaml')

        self.test_two_experiments_with_cs_args(
            config=config_path,
            cloud_timeout=100,
            cloud='ec2')

    @unittest.skipIf(not has_aws_credentials(),
                     'AWS credentials needed for this test')
    def test_two_experiments_ec2spot(self):
        mypath = os.path.dirname(os.path.realpath(__file__))
        config_path = os.path.join(
            mypath,
            '..',
            'tests',
            'test_config_http_client.yaml')

        self.test_two_experiments_with_cs_args(
            config=config_path,
            cloud_timeout=100,
            cloud='ec2spot')

    def test_two_experiments_apiserver(self):
        mypath = os.path.dirname(os.path.realpath(__file__))
        config_path = os.path.join(
            mypath,
            '..',
            'tests',
            'test_config_http_client.yaml')

        with get_local_queue_lock():
            self.test_two_experiments_with_cs_args(config=config_path)

    @unittest.skipIf(
        'GOOGLE_APPLICATION_CREDENTIALS' not in os.environ.keys(),
        'Need GOOGLE_APPLICATION_CREDENTIALS env variable to' +
        'use google cloud')
    def test_two_experiments_gcloud(self):
        mypath = os.path.dirname(os.path.realpath(__file__))
        config_path = os.path.join(
            mypath,
            '..',
            'tests',
            'test_config_http_client.yaml')

        self.test_two_experiments_with_cs_args(
            config=config_path,
            cloud='gcloud')

    @unittest.skip('TODO peterz scale down or fix')
    # @unittest.skipIf(not has_aws_credentials(),
    #                 'AWS credentials needed for this test')
    def test_many_experiments_ec2(self):
        experimentId = str(uuid.uuid4())
        mypath = os.path.dirname(os.path.realpath(__file__))
        config_path = os.path.join(
            mypath,
            '..',
            'tests',
            'test_config.yaml')

        n_experiments = 100
        num_workers = 30

        print("Executing {} tasks with {} workers"
              .format(n_experiments, num_workers))

        results = {}
        expected_results = {}

        logger = logging.getLogger('test_1k_experiments_ec2')
        logger.setLevel(10)

        with CompletionService(experimentId,
                               config=config_path, cloud='ec2spot',
                               num_workers=num_workers) as cs:

            def submit_task(i):
                key = cs.submitTaskWithFiles(
                    os.path.join(
                        mypath,
                        'completion_service_func.py'),
                    [i],
                    {
                        'a': '/Users/peter.zhokhov/.bash_profile',
                        'p': '/Users/peter.zhokhov/.bash_profile'
                    })
                logger.info('Submitted task ' + str(i))
                expected_results[key] = [i]

            '''
            pool.map(submit_task, range(n_experiments))
            print("Submitted")
            pool.close()
            pool.join()
            '''
            for i in range(n_experiments):
                submit_task(i)

            for i in range(0, n_experiments):
                print("Trying to get a result " + str(i))
                result = cs.getResults(blocking=True)
                logger.info('Received result ' + str(result))
                results[result[0]] = result[1]

        self.assertEquals(results, expected_results)

    @unittest.skipIf(
        'GOOGLE_APPLICATION_CREDENTIALS' not in os.environ.keys(),
        'Need GOOGLE_APPLICATION_CREDENTIALS env variable to' +
        'use google cloud')
    def test_two_experiments_gcloud_nonspot(self):
        mypath = os.path.dirname(os.path.realpath(__file__))
        config_path = os.path.join(
            mypath,
            '..',
            'tests',
            'test_config.yaml')

        self.test_two_experiments_with_cs_args(
            config=config_path,
            cloud='gcloud')


if __name__ == '__main__':
    unittest.main()



import unittest
import subprocess
import time
from random import randint
import os
import tempfile
import uuid

from studio import model
from model_test import get_test_experiment


@unittest.skipIf('GOOGLE_APPLICATION_CREDENTIALS' not in os.environ.keys(),
                 "GOOGLE_APPLICATION_CREDENTIALS is missing, needed for " +
                 "server to communicate with storage")
class HTTPProviderTest(unittest.TestCase):

    _mutliprocess_shared_ = True

    @classmethod
    def setUpClass(self):
        print("Starting up the API server")
        self.port = randint(5000, 9000)

        # self.app.run(port=self.port, debug=True)
        # self.serverp.start()
        self.server_config_file = os.path.join(
            os.path.dirname(
                os.path.realpath(__file__)),
            'test_config_http_server.yaml')

        print(self.server_config_file)

        self.client_config_file = os.path.join(
            os.path.dirname(
                os.path.realpath(__file__)),
            'test_config_http_client.yaml')

        self.serverp = subprocess.Popen([
            'studio-ui',
            '--port=' + str(self.port),
            '--verbose=debug',
            '--config=' + self.server_config_file,
            '--host=localhost'])

        time.sleep(25)

    @classmethod
    def tearDownClass(self):
        print("Shutting down the API server")
        self.serverp.kill()

    def get_db_provider(self):
        config = model.get_config(self.client_config_file)
        config['database']['serverUrl'] = 'http://localhost:' + str(self.port)
        return model.get_db_provider(config)

    def test_add_get_experiment(self):
        experiment_tuple = get_test_experiment()
        db = self.get_db_provider()
        db.add_experiment(experiment_tuple[0])

        experiment = db.get_experiment(experiment_tuple[0].key)
        self.assertEquals(experiment.key, experiment_tuple[0].key)
        self.assertEquals(experiment.filename, experiment_tuple[0].filename)
        self.assertEquals(experiment.args, experiment_tuple[0].args)

        db.delete_experiment(experiment_tuple[1])

    def test_add_get_experiment_artifacts(self):
        experiment_tuple = get_test_experiment()
        e_experiment = experiment_tuple[0]
        e_artifacts = e_experiment.artifacts

        a1_filename = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        a2_filename = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))

        with open(a1_filename, 'w') as f:
            f.write('hello world')

        e_artifacts['a1'] = {
            'local': a1_filename,
            'mutable': False
        }

        e_artifacts['a2'] = {
            'local': a2_filename,
            'mutable': True
        }

        db = self.get_db_provider()
        db.add_experiment(e_experiment)

        experiment = db.get_experiment(e_experiment.key)
        self.assertEquals(experiment.key, e_experiment.key)
        self.assertEquals(experiment.filename, e_experiment.filename)
        self.assertEquals(experiment.args, e_experiment.args)
        db.delete_experiment(e_experiment.key)
        os.remove(a1_filename)

    def test_start_experiment(self):
        db = self.get_db_provider()
        experiment_tuple = get_test_experiment()

        db.add_experiment(experiment_tuple[0])
        db.start_experiment(experiment_tuple[0])

        experiment = db.get_experiment(experiment_tuple[1])

        self.assertTrue(experiment.status == 'running')
        self.assertTrue(experiment.time_added <= time.time())
        self.assertTrue(experiment.time_started <= time.time())

        self.assertEquals(experiment.key, experiment_tuple[0].key)
        self.assertEquals(experiment.filename, experiment_tuple[0].filename)
        self.assertEquals(experiment.args, experiment_tuple[0].args)

        db.finish_experiment(experiment_tuple[0])
        db.delete_experiment(experiment_tuple[1])


if __name__ == '__main__':
    unittest.main()



import unittest
import os
import shutil
import tempfile
import uuid
import subprocess
import time
from timeout_decorator import timeout
import logging
import traceback
import numpy as np

try:
    import keras
except BaseException:
    keras = None

from studio import model
from studio.local_queue import LocalQueue, get_local_queue_lock
from studio.util import has_aws_credentials, sixdecode, retry

logging.basicConfig()


class LocalWorkerTest(unittest.TestCase):

    @timeout(500, use_signals=False)
    def test_runner_local(self):
        with get_local_queue_lock():
            with stubtest_worker(
                self,
                experiment_name='test_runner_local_' + str(uuid.uuid4()),
                runner_args=['--verbose=debug'],
                config_name='test_config_http_client.yaml',
                test_script='tf_hello_world.py',
                script_args=['arg0'],
                expected_output='[ 2.  6.]'
            ):
                pass

    @timeout(500, use_signals=False)
    def test_args_conflict(self):
        with get_local_queue_lock():
            with stubtest_worker(
                self,
                experiment_name='test_runner_conflict_' + str(uuid.uuid4()),
                runner_args=['--verbose=debug'],
                config_name='test_config.yaml',
                test_script='conflicting_args.py',
                script_args=['--experiment', 'aaa'],
                expected_output='Experiment key = aaa'
            ):
                pass

    @timeout(500, use_signals=False)
    def test_local_hyperparam(self):
        with get_local_queue_lock():
            with stubtest_worker(
                self,
                experiment_name='test_local_hyperparam' + str(uuid.uuid4()),
                runner_args=['--verbose=debug'],
                config_name='test_config_http_client.yaml',
                test_script='hyperparam_hello_world.py',
                expected_output='0.3'
            ):
                pass

            with stubtest_worker(
                self,
                experiment_name='test_local_hyperparam' + str(uuid.uuid4()),
                runner_args=[
                    '--verbose=debug',
                    '--hyperparam=learning_rate=0.4'
                ],
                config_name='test_config_http_client.yaml',
                test_script='hyperparam_hello_world.py',
                expected_output='0.4'
            ):
                pass

    @timeout(500, use_signals=False)
    def test_local_worker_ce(self):
        tmpfile = os.path.join(tempfile.gettempdir(),
                               'tmpfile' +
                               str(uuid.uuid4()) + '.txt')

        random_str1 = str(uuid.uuid4())
        with open(tmpfile, 'w') as f:
            f.write(random_str1)

        random_str2 = str(uuid.uuid4())
        experiment_name = 'test_local_worker_c' + str(uuid.uuid4())

        with get_local_queue_lock():
            with stubtest_worker(
                self,
                experiment_name=experiment_name,
                runner_args=['--capture=' + tmpfile + ':f',
                             '--verbose=debug'],
                config_name='test_config_http_client.yaml',
                test_script='art_hello_world.py',
                script_args=[random_str2],
                expected_output=random_str1,
                delete_when_done=False
            ) as db:

                tmppath = os.path.join(
                    tempfile.gettempdir(), str(
                        uuid.uuid4()))

                db.get_artifact(
                    db.get_experiment(experiment_name).artifacts['f'],
                    tmppath)

                with open(tmppath, 'r') as f:
                    self.assertTrue(f.read() == random_str2)
                os.remove(tmppath)

            with stubtest_worker(
                self,
                experiment_name='test_local_worker_e' + str(uuid.uuid4()),
                runner_args=['--reuse={}/f:f'.format(experiment_name)],
                config_name='test_config_http_client.yaml',
                test_script='art_hello_world.py',
                script_args=[],
                expected_output=random_str2
            ) as db:

                db.delete_experiment(experiment_name)

    @timeout(500, use_signals=False)
    def test_local_worker_co(self):
        tmpfile = os.path.join(tempfile.gettempdir(),
                               'tmpfile' +
                               str(uuid.uuid4()) + '.txt')

        random_str = str(uuid.uuid4())
        with open(tmpfile, 'w') as f:
            f.write(random_str)

        with get_local_queue_lock():
            with stubtest_worker(
                self,
                experiment_name='test_local_worker_co' + str(uuid.uuid4()),
                runner_args=['--capture-once=' + tmpfile + ':f'],
                config_name='test_config_http_client.yaml',
                test_script='art_hello_world.py',
                script_args=[],
                expected_output=random_str
            ):
                pass

    @timeout(500, use_signals=False)
    def test_local_worker_co_url(self):
        expected_str = 'Zabil zaryad ya v pushku tugo'
        url = 'https://storage.googleapis.com/studio-ed756.appspot.com/' + \
              'tests/url_artifact.txt'

        with get_local_queue_lock():
            with stubtest_worker(
                self,
                experiment_name='test_local_worker_co_url' + str(uuid.uuid4()),
                runner_args=['--capture-once=' + url + ':f'],
                config_name='test_config_http_client.yaml',
                test_script='art_hello_world.py',
                script_args=[],
                expected_output=expected_str
            ):
                pass

    @unittest.skipIf(
        not has_aws_credentials(),
        'AWS credentials not found, cannot download s3://-like links')
    @timeout(500, use_signals=False)
    def test_local_worker_co_s3(self):
        expected_str = 'No4 ulica fonar apteka, bessmyslennyj i tusklyj svet'
        s3loc = 's3://studioml-artifacts/tests/download_test/download_test.txt'

        with get_local_queue_lock():
            with stubtest_worker(
                self,
                experiment_name='test_local_worker_co_s3' + str(uuid.uuid4()),
                runner_args=['--capture-once=' + s3loc + ':f'],
                config_name='test_config_http_client.yaml',
                test_script='art_hello_world.py',
                script_args=[],
                expected_output=expected_str
            ):
                pass

    @unittest.skipIf(keras is None,
                     'keras is required for this test')
    @timeout(500, use_signals=False)
    def test_save_get_model(self):
        experiment_name = 'test_save_get_model' + str(uuid.uuid4())
        with get_local_queue_lock():
            with stubtest_worker(
                self,
                experiment_name=experiment_name,
                runner_args=[],
                config_name='test_config_http_client.yaml',
                test_script='save_model.py',
                script_args=[],
                expected_output='',
                delete_when_done=False,
                test_output=False
            ) as db:

                experiment = db.get_experiment(experiment_name)
                saved_model = experiment.get_model(db)

                v = np.random.rand(1, 2)
                prediction = saved_model.predict(v)
                expected = v * 2

                self.assertTrue(np.isclose(prediction, expected).all())

                db.delete_experiment(experiment)

    @timeout(500, use_signals=False)
    def test_stop_experiment(self):
        my_path = os.path.dirname(os.path.realpath(__file__))

        logger = logging.getLogger('test_stop_experiment')
        logger.setLevel(10)

        config_name = os.path.join(my_path, 'test_config_http_client.yaml')
        key = 'test_stop_experiment' + str(uuid.uuid4())

        with model.get_db_provider(model.get_config(config_name)) as db:
            try:
                db.delete_experiment(key)
            except Exception:
                pass

            with get_local_queue_lock():
                p = subprocess.Popen(['studio', 'run',
                                      '--config=' + config_name,
                                      '--experiment=' + key,
                                      '--force-git',
                                      '--verbose=debug',
                                      'stop_experiment.py'],
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.STDOUT,
                                     cwd=my_path)

                # wait till experiment spins up
                experiment = None
                while experiment is None or experiment.status == 'waiting':
                    time.sleep(1)
                    try:
                        experiment = db.get_experiment(key)
                    except BaseException:
                        pass

                logger.info('Stopping experiment')
                db.stop_experiment(key)
                pout, _ = p.communicate()

            if pout:
                logger.debug("studio run output: \n" + pout.decode())

            db.delete_experiment(key)

    @timeout(500, use_signals=False)
    def test_experiment_maxduration(self):
        my_path = os.path.dirname(os.path.realpath(__file__))

        logger = logging.getLogger('test_experiment_maxduration')
        logger.setLevel(10)

        config_name = os.path.join(my_path, 'test_config_http_client.yaml')
        key = 'test_experiment_maxduration' + str(uuid.uuid4())

        with model.get_db_provider(model.get_config(config_name)) as db:
            try:
                db.delete_experiment(key)
            except Exception:
                pass

            with get_local_queue_lock():
                p = subprocess.Popen(['studio', 'run',
                                      '--config=' + config_name,
                                      '--experiment=' + key,
                                      '--force-git',
                                      '--verbose=debug',
                                      '--max-duration=10s',
                                      'stop_experiment.py'],
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.STDOUT,
                                     cwd=my_path)

                pout, _ = p.communicate()
            if pout:
                logger.debug("studio run output: \n" + pout.decode())

            db.delete_experiment(key)

    @timeout(500, use_signals=False)
    def test_experiment_lifetime(self):
        my_path = os.path.dirname(os.path.realpath(__file__))

        logger = logging.getLogger('test_experiment_lifetime')
        logger.setLevel(10)

        config_name = os.path.join(my_path, 'test_config_http_client.yaml')
        key = 'test_experiment_lifetime' + str(uuid.uuid4())

        with model.get_db_provider(model.get_config(config_name)) as db:
            try:
                db.delete_experiment(key)
            except Exception:
                pass

            with get_local_queue_lock():
                p = subprocess.Popen(['studio', 'run',
                                      '--config=' + config_name,
                                      '--experiment=' + key,
                                      '--force-git',
                                      '--verbose=debug',
                                      '--lifetime=-10m',
                                      'stop_experiment.py'],
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.STDOUT,
                                     cwd=my_path)

                pout, _ = p.communicate()

            if pout:
                logger.debug("studio run output: \n" + pout.decode())

            db.delete_experiment(key)


def stubtest_worker(
        testclass,
        experiment_name,
        runner_args,
        config_name,
        test_script,
        expected_output,
        script_args=[],
        queue=LocalQueue(),
        wait_for_experiment=True,
        delete_when_done=True,
        test_output=True,
        test_workspace=True):

    my_path = os.path.dirname(os.path.realpath(__file__))
    config_name = os.path.join(my_path, config_name)
    logger = logging.getLogger('stubtest_worker')
    logger.setLevel(10)

    queue.clean()

    with model.get_db_provider(model.get_config(config_name)) as db:
        try:
            db.delete_experiment(experiment_name)
        except Exception:
            pass

    p = subprocess.Popen(['studio', 'run'] + runner_args +
                         ['--config=' + config_name,
                          '--verbose=debug',
                          '--force-git',
                          '--experiment=' + experiment_name,
                          test_script] + script_args,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.STDOUT,
                         close_fds=True,
                         cwd=my_path)

    pout, _ = p.communicate()

    if pout:
        logger.debug("studio run output: \n" + sixdecode(pout))

    db = model.get_db_provider(model.get_config(config_name))
    experiments = [e for e in db.get_user_experiments()
                   if e.startswith(experiment_name)]

    assert len(experiments) == 1, "actually {} number of experiments".format(
        len(experiments))

    experiment_name = experiments[0]

    try:
        experiment = db.get_experiment(experiment_name)
        if wait_for_experiment:
            while not experiment or not experiment.status == 'finished':
                experiment = db.get_experiment(experiment_name)

        if test_output:
            with open(db.get_artifact(experiment.artifacts['output']),
                      'r') as f:
                data = f.read()
                split_data = data.strip().split('\n')
                testclass.assertEquals(split_data[-1], expected_output)

        if test_workspace:
            check_workspace(testclass, db, experiment_name)

        if delete_when_done:
            retry(lambda: db.delete_experiment(experiment_name), sleep_time=10)

        return db

    except Exception as e:
        print("Exception {} raised during test".format(e))
        print("worker output: \n {}".format(pout))
        print("Exception trace:")
        print(traceback.format_exc())
        raise e


def check_workspace(testclass, db, key):

    tmpdir = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
    os.mkdir(tmpdir)
    artifact = db.get_experiment(key).artifacts['workspace']
    localpath = db.get_artifact(artifact,
                                tmpdir, only_newer=False)

    for _, _, files in os.walk(localpath, topdown=False):
        for filename in files:
            downloaded_filename = os.path.join(tmpdir, filename)
            if downloaded_filename.endswith('.pyc'):
                continue

            with open(downloaded_filename, 'rb') as f1:
                data1 = f1.read()
            with open(os.path.join(db.get_artifact(artifact), filename),
                      'rb') as f2:
                data2 = f2.read()

            testclass.assertTrue(
                data1 == data2,
                ('File comparison between local {} ' +
                 'and downloaded {} has failed')
                .format(
                    filename,
                    downloaded_filename))

    for _, _, files in os.walk('tmpdir', topdown=False):
        for filename in files:
            downloaded_filename = os.path.join(tmpdir, filename)
            with open(downloaded_filename, 'rb') as f1:
                data1 = f1.read()
            with open(filename, 'rb') as f2:
                data2 = f2.read()

            testclass.assertTrue(data1 == data2)

    shutil.rmtree(tmpdir)


if __name__ == "__main__":
    unittest.main()



import sys

from keras.layers import Dense, Flatten

from keras.models import Sequential
from keras.datasets import mnist
from keras.utils import to_categorical

from keras.callbacks import ModelCheckpoint, TensorBoard
from keras import optimizers

from studio import fs_tracker


(x_train, y_train), (x_test, y_test) = mnist.load_data()

x_train = x_train.reshape(60000, 28, 28, 1)
x_test = x_test.reshape(10000, 28, 28, 1)
x_train = x_train.astype('float32')
x_test = x_test.astype('float32')
x_train /= 255
x_test /= 255

# convert class vectors to binary class matrices
y_train = to_categorical(y_train, 10)
y_test = to_categorical(y_test, 10)


model = Sequential()

model.add(Flatten(input_shape=(28, 28, 1)))
model.add(Dense(128, activation='relu'))
model.add(Dense(128, activation='relu'))

model.add(Dense(10, activation='softmax'))
model.summary()


batch_size = 128
no_epochs = int(sys.argv[1]) if len(sys.argv) > 1 else 10
lr = 0.01

print('learning rate = {}'.format(lr))
print('batch size = {}'.format(batch_size))
print('no_epochs = {}'.format(no_epochs))

model.compile(loss='categorical_crossentropy', optimizer=optimizers.SGD(lr=lr),
              metrics=['accuracy'])


checkpointer = ModelCheckpoint(
    fs_tracker.get_model_directory() +
    '/checkpoint.{epoch:02d}-{val_loss:.2f}.hdf')


tbcallback = TensorBoard(log_dir=fs_tracker.get_tensorboard_dir(),
                         histogram_freq=0,
                         write_graph=True,
                         write_images=True)


model.fit(
    x_train, y_train, validation_data=(
        x_test,
        y_test),
    epochs=no_epochs,
    callbacks=[checkpointer, tbcallback],
    batch_size=batch_size)
